from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path


_ENV_PRIORITY = (".env.local", ".env.production", ".env.development", ".env.test")

_ENV_PREFIX = "GENAI_CALLING_"


def get_prefixed_env(name: str) -> str | None:
    for env_name in prefixed_env_names(name):
        value = os.environ.get(env_name)
        if value is not None:
            return value
    return None


def get_prefixed_env_int(name: str, default: int) -> int:
    raw = get_prefixed_env(name)
    if raw is None:
        return default
    try:
        return int(raw)
    except ValueError:
        return default


def prefixed_env_names(name: str) -> tuple[str, ...]:
    return (f"{_ENV_PREFIX}{name}",)


def format_prefixed_env_names(name: str, *extra_names: str) -> str:
    names = [*prefixed_env_names(name), *extra_names]
    return "/".join(names)


def _parse_env_line(line: str) -> tuple[str, str] | None:
    stripped = line.strip()
    if not stripped or stripped.startswith("#"):
        return None
    if "=" not in stripped:
        return None
    key, value = stripped.split("=", 1)
    key = key.strip()
    value = value.strip()
    if not key:
        return None
    if (value.startswith("'") and value.endswith("'")) or (
        value.startswith('"') and value.endswith('"')
    ):
        value = value[1:-1]
    return key, value


def _global_env_paths() -> list[Path]:
    return [Path.home() / ".genai-calling" / ".env"]


def load_env_files(root: str | Path | None = None) -> list[Path]:
    """
    Load env files by priority:
    `.env.local > .env.production > .env.development > .env.test >
    ~/.genai-calling/.env`.

    Implementation: apply higher priority first without overriding existing env.
    """
    base = Path(root) if root is not None else Path.cwd()
    loaded: list[Path] = []
    paths = [base / name for name in _ENV_PRIORITY]
    paths.extend(_global_env_paths())
    for path in paths:
        if not path.is_file():
            continue
        loaded.append(path)
        for line in path.read_text(encoding="utf-8").splitlines():
            parsed = _parse_env_line(line)
            if parsed is None:
                continue
            key, value = parsed
            os.environ.setdefault(key, value)
    return loaded


@dataclass(frozen=True, slots=True)
class ProviderKeys:
    openai_api_key: str | None
    google_api_key: str | None
    anthropic_api_key: str | None
    aliyun_api_key: str | None
    volcengine_api_key: str | None
    tuzi_web_api_key: str | None
    tuzi_openai_api_key: str | None
    tuzi_google_api_key: str | None
    tuzi_anthropic_api_key: str | None


def get_provider_keys() -> ProviderKeys:
    return ProviderKeys(
        openai_api_key=get_prefixed_env("OPENAI_API_KEY")
        or os.environ.get("OPENAI_API_KEY"),
        google_api_key=get_prefixed_env("GOOGLE_API_KEY")
        or os.environ.get("GOOGLE_API_KEY"),
        anthropic_api_key=get_prefixed_env("ANTHROPIC_API_KEY")
        or os.environ.get("ANTHROPIC_API_KEY"),
        aliyun_api_key=get_prefixed_env("ALIYUN_API_KEY")
        or os.environ.get("ALIYUN_API_KEY"),
        volcengine_api_key=get_prefixed_env("VOLCENGINE_API_KEY")
        or os.environ.get("VOLCENGINE_API_KEY"),
        tuzi_web_api_key=get_prefixed_env("TUZI_WEB_API_KEY")
        or os.environ.get("TUZI_WEB_API_KEY"),
        tuzi_openai_api_key=get_prefixed_env("TUZI_OPENAI_API_KEY")
        or os.environ.get("TUZI_OPENAI_API_KEY"),
        tuzi_google_api_key=get_prefixed_env("TUZI_GOOGLE_API_KEY")
        or os.environ.get("TUZI_GOOGLE_API_KEY"),
        tuzi_anthropic_api_key=get_prefixed_env("TUZI_ANTHROPIC_API_KEY")
        or os.environ.get("TUZI_ANTHROPIC_API_KEY"),
    )


def get_default_timeout_ms() -> int:
    raw = get_prefixed_env("TIMEOUT_MS")
    if raw is None:
        return 120_000
    try:
        value = int(raw)
    except ValueError:
        return 120_000
    return max(1, value)



import unittest


class TestCapabilityRules(unittest.TestCase):
    def test_openai_adapter_modalities_match_series_rules(self) -> None:
        from gravtice.genai.providers.openai import OpenAIAdapter

        adapter = OpenAIAdapter(api_key="__demo__")

        self.assertEqual(adapter.capabilities("deepseek-v3").input_modalities, {"text"})
        self.assertEqual(adapter.capabilities("kimi-k2").input_modalities, {"text"})
        self.assertEqual(adapter.capabilities("qvq-72b").input_modalities, {"text"})

        self.assertEqual(
            adapter.capabilities("kimi-latest").input_modalities, {"text", "image"}
        )

        self.assertEqual(
            adapter.capabilities("qwen2.5-vl-72b-instruct").input_modalities,
            {"text", "image"},
        )
        self.assertEqual(
            adapter.capabilities("qwen2.5-72b-instruct").input_modalities, {"text"}
        )

        cap_asr = adapter.capabilities("qwen-asr")
        self.assertEqual(cap_asr.input_modalities, {"text", "audio"})
        self.assertEqual(cap_asr.output_modalities, {"text"})

        cap_z_image = adapter.capabilities("z-image-1")
        self.assertEqual(cap_z_image.input_modalities, {"text", "image"})
        self.assertEqual(cap_z_image.output_modalities, {"image"})

    def test_gemini_adapter_modalities_match_series_rules(self) -> None:
        from gravtice.genai.providers.gemini import GeminiAdapter

        adapter = GeminiAdapter(
            api_key="__demo__",
            base_url="https://example.invalid",
            provider_name="google",
        )

        cap_imagen = adapter.capabilities("imagen-4.0-generate-001")
        self.assertEqual(cap_imagen.input_modalities, {"text"})
        self.assertEqual(cap_imagen.output_modalities, {"image"})

        cap_gemini_image = adapter.capabilities("gemini-2.5-flash-image")
        self.assertEqual(cap_gemini_image.input_modalities, {"text", "image"})
        self.assertEqual(cap_gemini_image.output_modalities, {"image"})

        cap_native_audio = adapter.capabilities("gemini-2.5-flash-native-audio-latest")
        self.assertEqual(cap_native_audio.input_modalities, {"text", "audio", "video"})
        self.assertEqual(cap_native_audio.output_modalities, {"text", "audio"})

        cap_veo = adapter.capabilities("veo-3.0-generate")
        self.assertEqual(cap_veo.input_modalities, {"text"})
        self.assertEqual(cap_veo.output_modalities, {"video"})

        cap_tts = adapter.capabilities("gemini-2.5-pro-preview-tts")
        self.assertEqual(cap_tts.input_modalities, {"text"})
        self.assertEqual(cap_tts.output_modalities, {"audio"})



import unittest


class _FakeSseResponse:
    def __init__(self, lines: list[bytes]) -> None:
        self._lines = list(lines)
        self._idx = 0

    def readline(self) -> bytes:
        if self._idx >= len(self._lines):
            return b""
        line = self._lines[self._idx]
        self._idx += 1
        return line


class TestSseParser(unittest.TestCase):
    def test_parses_event_id_retry_and_multiline_data(self) -> None:
        from gravtice.genai._internal.http import SSEEvent, _iter_sse_events

        resp = _FakeSseResponse(
            [
                b": keepalive\n",
                b"event: message\n",
                b"id: abc123\n",
                b"retry: 5000\n",
                b'data: {"a":1}\n',
                b'data: {"b":2}\n',
                b"\n",
            ]
        )
        events = list(_iter_sse_events(resp))
        self.assertEqual(
            events,
            [
                SSEEvent(
                    data='{"a":1}\n{"b":2}',
                    event="message",
                    id="abc123",
                    retry=5000,
                )
            ],
        )

    def test_id_with_null_is_ignored(self) -> None:
        from gravtice.genai._internal.http import SSEEvent, _iter_sse_events

        resp = _FakeSseResponse(
            [
                b"id: ok\x00nope\n",
                b"data: {}\n",
                b"\n",
            ]
        )
        events = list(_iter_sse_events(resp))
        self.assertEqual(events, [SSEEvent(data="{}", event=None, id=None, retry=None)])

    def test_event_without_data_is_emitted_with_empty_data(self) -> None:
        from gravtice.genai._internal.http import SSEEvent, _iter_sse_events

        resp = _FakeSseResponse(
            [
                b"id: 1\n",
                b"\n",
            ]
        )
        events = list(_iter_sse_events(resp))
        self.assertEqual(events, [SSEEvent(data="", event=None, id="1", retry=None)])



from __future__ import annotations

import base64
import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Literal, cast

from ._internal.errors import ErrorInfo, invalid_request_error, not_supported_error

Role = Literal["system", "user", "assistant", "tool"]
PartType = Literal[
    "text",
    "image",
    "audio",
    "video",
    "embedding",
    "file",
    "tool_call",
    "tool_result",
]

Modality = Literal["text", "image", "audio", "video", "embedding"]
Status = Literal["completed", "running", "failed", "canceled"]
ReasoningEffort = Literal["none", "minimal", "low", "medium", "high", "xhigh"]
ToolChoiceMode = Literal["none", "auto", "required", "tool"]


@dataclass(frozen=True, slots=True)
class ReasoningSpec:
    effort: ReasoningEffort | None = None


_REASONING_EFFORT_VALUES = {"none", "minimal", "low", "medium", "high", "xhigh"}


def normalize_reasoning_effort(value: object) -> ReasoningEffort:
    if not isinstance(value, str) or not value.strip():
        raise invalid_request_error("reasoning.effort must be a non-empty string")
    effort = value.strip().lower()
    if effort not in _REASONING_EFFORT_VALUES:
        raise invalid_request_error(f"unknown reasoning.effort: {effort}")
    return cast(ReasoningEffort, effort)


@dataclass(frozen=True, slots=True)
class PartSourceBytes:
    kind: Literal["bytes"] = "bytes"
    data: bytes | str = b""
    encoding: Literal["base64"] | None = None

    def __post_init__(self) -> None:
        if self.encoding is None:
            if isinstance(self.data, bytearray):
                object.__setattr__(self, "data", bytes(self.data))
                return
            if not isinstance(self.data, bytes):
                raise ValueError(
                    "PartSourceBytes.data must be bytes when encoding is None"
                )
            return

        if self.encoding != "base64":
            raise ValueError(f"unknown PartSourceBytes.encoding: {self.encoding}")
        if not isinstance(self.data, str):
            raise ValueError(
                "PartSourceBytes.data must be str when encoding is 'base64'"
            )


@dataclass(frozen=True, slots=True)
class PartSourcePath:
    kind: Literal["path"] = "path"
    path: str = ""


@dataclass(frozen=True, slots=True)
class PartSourceUrl:
    kind: Literal["url"] = "url"
    url: str = ""


@dataclass(frozen=True, slots=True)
class PartSourceRef:
    kind: Literal["ref"] = "ref"
    provider: str = ""
    id: str = ""


PartSource = PartSourceBytes | PartSourcePath | PartSourceUrl | PartSourceRef


@dataclass(frozen=True, slots=True)
class Part:
    type: PartType
    mime_type: str | None = None
    source: PartSource | None = None
    text: str | None = None
    embedding: list[float] | None = None
    meta: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        if not isinstance(self.meta, dict):
            raise ValueError("Part.meta must be an object")
        if self.mime_type is not None and not isinstance(self.mime_type, str):
            raise ValueError("Part.mime_type must be a string")

        if self.type == "text":
            if not isinstance(self.text, str):
                raise ValueError("text Part requires text")
            if self.source is not None:
                raise ValueError("text Part cannot have source")
            if self.embedding is not None:
                raise ValueError("text Part cannot have embedding")
            return

        if self.type == "embedding":
            if not isinstance(self.embedding, list) or not all(
                isinstance(x, (int, float)) for x in self.embedding
            ):
                raise ValueError("embedding Part requires embedding: list[number]")
            if self.source is not None:
                raise ValueError("embedding Part cannot have source")
            if self.text is not None:
                raise ValueError("embedding Part cannot have text")
            return

        if self.type in {"image", "audio", "video", "file"}:
            if self.source is None:
                raise ValueError(f"{self.type} Part requires source")
            if not isinstance(
                self.source,
                (PartSourceBytes, PartSourcePath, PartSourceUrl, PartSourceRef),
            ):
                raise ValueError("Part.source must be a PartSource object")
            if self.text is not None:
                raise ValueError(f"{self.type} Part cannot have text")
            if self.embedding is not None:
                raise ValueError(f"{self.type} Part cannot have embedding")
            if self.mime_type and self.type in {"image", "audio", "video"}:
                prefix = f"{self.type}/"
                if not self.mime_type.startswith(prefix):
                    raise ValueError(
                        f"{self.type} Part mime_type must start with {prefix!r}"
                    )
            return

        if self.type in {"tool_call", "tool_result"}:
            if self.source is not None:
                raise ValueError(f"{self.type} Part cannot have source")
            if self.text is not None:
                raise ValueError(f"{self.type} Part cannot have text")
            if self.embedding is not None:
                raise ValueError(f"{self.type} Part cannot have embedding")
            return

        raise ValueError(f"unknown Part.type: {self.type}")

    @staticmethod
    def from_text(text: str) -> "Part":
        return Part(type="text", text=text)

    @staticmethod
    def tool_call(
        *, name: str, arguments: Any, tool_call_id: str | None = None
    ) -> "Part":
        meta: dict[str, Any] = {"name": name, "arguments": arguments}
        if tool_call_id is not None:
            meta["tool_call_id"] = tool_call_id
        return Part(type="tool_call", meta=meta)

    @staticmethod
    def tool_result(
        *,
        name: str,
        result: Any,
        tool_call_id: str | None = None,
        is_error: bool | None = None,
    ) -> "Part":
        meta: dict[str, Any] = {"name": name, "result": result}
        if tool_call_id is not None:
            meta["tool_call_id"] = tool_call_id
        if is_error is not None:
            meta["is_error"] = bool(is_error)
        return Part(type="tool_result", meta=meta)

    @staticmethod
    def embedding_part(vector: list[float]) -> "Part":
        return Part(type="embedding", embedding=vector)

    def require_text(self) -> str:
        if self.type != "text" or self.text is None:
            raise invalid_request_error("Part is not text")
        return self.text

    def require_source(self) -> PartSource:
        if self.source is None:
            raise invalid_request_error("Part has no source")
        return self.source


@dataclass(frozen=True, slots=True)
class Message:
    role: Role
    content: list[Part]


@dataclass(frozen=True, slots=True)
class OutputTextSpec:
    format: Literal["text", "json"] = "text"
    json_schema: Any | None = None
    max_output_tokens: int | None = None


@dataclass(frozen=True, slots=True)
class OutputImageSpec:
    n: int | None = None
    size: str | None = None
    format: str | None = None


@dataclass(frozen=True, slots=True)
class OutputAudioSpec:
    voice: str | None = None
    format: str | None = None
    language: str | None = None


@dataclass(frozen=True, slots=True)
class OutputVideoSpec:
    duration_sec: int | None = None
    aspect_ratio: str | None = None
    fps: int | None = None
    format: str | None = None


@dataclass(frozen=True, slots=True)
class OutputEmbeddingSpec:
    dimensions: int | None = None


@dataclass(frozen=True, slots=True)
class Tool:
    """
    Minimal function tool declaration (provider-agnostic).

    - `parameters` is a JSON Schema object describing the function arguments.
    - Providers may support only a subset of JSON Schema/OpenAPI Schema.
    """

    name: str
    description: str | None = None
    parameters: dict[str, Any] | None = None
    strict: bool | None = None


@dataclass(frozen=True, slots=True)
class ToolChoice:
    mode: ToolChoiceMode = "auto"
    name: str | None = None

    def normalized(self) -> "ToolChoice":
        mode = self.mode.strip().lower()
        if mode not in {"none", "auto", "required", "tool"}:
            raise invalid_request_error(f"unknown tool_choice.mode: {self.mode}")
        name = self.name.strip() if isinstance(self.name, str) else None
        if mode == "tool" and not name:
            raise invalid_request_error("tool_choice.name required when mode='tool'")
        if mode != "tool" and name is not None:
            raise invalid_request_error(
                "tool_choice.name only allowed when mode='tool'"
            )
        return ToolChoice(mode=cast(ToolChoiceMode, mode), name=name)


@dataclass(frozen=True, slots=True)
class OutputSpec:
    modalities: list[Modality]
    text: OutputTextSpec | None = None
    image: OutputImageSpec | None = None
    audio: OutputAudioSpec | None = None
    video: OutputVideoSpec | None = None
    embedding: OutputEmbeddingSpec | None = None


@dataclass(frozen=True, slots=True)
class GenerateParams:
    temperature: float | None = None
    top_p: float | None = None
    seed: int | None = None
    max_output_tokens: int | None = None
    stop: list[str] | None = None
    timeout_ms: int | None = None
    idempotency_key: str | None = None
    reasoning: ReasoningSpec | None = None


@dataclass(frozen=True, slots=True)
class GenerateRequest:
    model: str = field(
        metadata={
            "description": 'Model string in the form "{provider}:{model_id}" (e.g. "openai:gpt-4o-mini")',
            "pattern": r"^[^\s:]+:[^\s]+$",
            "examples": ["openai:gpt-4o-mini"],
        }
    )
    input: list[Message]
    output: OutputSpec
    params: GenerateParams = field(default_factory=GenerateParams)
    wait: bool = True
    tools: list[Tool] | None = None
    tool_choice: ToolChoice | None = None
    provider_options: dict[str, Any] = field(default_factory=dict)

    def provider(self) -> str:
        if ":" not in self.model:
            raise invalid_request_error('model must be "{provider}:{model_id}"')
        return self.model.split(":", 1)[0]

    def model_id(self) -> str:
        if ":" not in self.model:
            raise invalid_request_error('model must be "{provider}:{model_id}"')
        return self.model.split(":", 1)[1]


@dataclass(frozen=True, slots=True)
class Usage:
    input_tokens: int | None = None
    output_tokens: int | None = None
    total_tokens: int | None = None
    seconds: float | None = None
    image_count: int | None = None
    video_seconds: float | None = None
    cost_estimate: float | None = None


@dataclass(frozen=True, slots=True)
class JobInfo:
    job_id: str
    poll_after_ms: int = 1_000
    expires_at: str | None = None
    last_status: str | None = None
    last_detail: str | None = None


@dataclass(frozen=True, slots=True)
class GenerateResponse:
    id: str
    provider: str
    model: str
    status: Status
    output: list[Message] = field(default_factory=list)
    usage: Usage | None = None
    job: JobInfo | None = None
    error: ErrorInfo | None = None


@dataclass(frozen=True, slots=True)
class GenerateEvent:
    type: str
    data: dict[str, Any] = field(default_factory=dict)


@dataclass(frozen=True, slots=True)
class Capability:
    input_modalities: set[Modality]
    output_modalities: set[Modality]
    supports_stream: bool
    supports_job: bool
    supports_tools: bool = False
    supports_json_schema: bool = False


def detect_mime_type(path: str) -> str | None:
    suffix = Path(path).suffix.lower()
    if suffix in {".png"}:
        return "image/png"
    if suffix in {".jpg", ".jpeg"}:
        return "image/jpeg"
    if suffix in {".webp"}:
        return "image/webp"
    if suffix in {".wav"}:
        return "audio/wav"
    if suffix in {".mp3"}:
        return "audio/mpeg"
    if suffix in {".m4a"}:
        return "audio/mp4"
    if suffix in {".mp4"}:
        return "video/mp4"
    if suffix in {".mov"}:
        return "video/quicktime"
    return None


def file_to_bytes(path: str, max_bytes: int) -> bytes:
    st = os.stat(path)
    if st.st_size > max_bytes:
        raise not_supported_error(
            f"file too large for inline bytes ({st.st_size} > {max_bytes}); use url/ref instead"
        )
    with open(path, "rb") as f:
        return f.read()


def bytes_to_base64(data: bytes) -> str:
    return base64.b64encode(data).decode("ascii")


def sniff_image_mime_type(data: bytes) -> str | None:
    if len(data) >= 8 and data.startswith(b"\x89PNG\r\n\x1a\n"):
        return "image/png"
    if len(data) >= 3 and data.startswith(b"\xff\xd8\xff"):
        return "image/jpeg"
    if len(data) >= 12 and data.startswith(b"RIFF") and data[8:12] == b"WEBP":
        return "image/webp"
    if len(data) >= 6 and data[:6] in {b"GIF87a", b"GIF89a"}:
        return "image/gif"
    return None



from __future__ import annotations

import argparse
import base64
import hashlib
import hmac
import json
import os
import time
from hmac import compare_digest
from collections import OrderedDict
from contextvars import ContextVar
from dataclasses import asdict, dataclass, replace
from typing import Any, cast
from typing_extensions import TypedDict
from urllib.parse import parse_qs
from uuid import uuid4

from ._internal.config import (
    format_prefixed_env_names,
    get_prefixed_env,
    get_prefixed_env_int,
)
from .client import Client
from .types import (
    GenerateRequest,
    GenerateResponse,
    OutputImageSpec,
    sniff_image_mime_type,
)

_MCP_GENERATE_REQUEST_SCHEMA: dict = {
    "type": "object",
    "title": "GenerateRequest",
    "description": 'genai-calling request object. `model` must be "{provider}:{model_id}" (e.g. "openai:gpt-4o-mini").',
    "required": ["model", "input", "output"],
    "properties": {
        "model": {
            "type": "string",
            "pattern": r"^[^\s:]+:[^\s]+$",
            "description": 'Model string in the form "{provider}:{model_id}".',
            "examples": ["openai:gpt-4o-mini"],
        },
        "input": {
            "description": 'Chat messages (or a shorthand prompt string). Preferred: [{"role":"user","content":[{"type":"text","text":"..."}]}]. Shorthand also accepted: "..." or [{"role":"user","content":"..."}].',
            "examples": [
                [
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": "Generate an image of a cute cat"}
                        ],
                    }
                ]
            ],
            "anyOf": [
                {"type": "string"},
                {
                    "type": "object",
                    "required": ["role", "content"],
                    "properties": {
                        "role": {
                            "type": "string",
                            "enum": ["system", "user", "assistant", "tool"],
                        },
                        "content": {
                            "description": 'List of Part objects (preferred) or a shorthand string. Preferred: [{"type":"text","text":"..."}].',
                            "anyOf": [
                                {"type": "string"},
                                {"type": "object"},
                                {"type": "array", "items": {"type": "object"}},
                            ],
                        },
                    },
                },
                {
                    "type": "array",
                    "minItems": 1,
                    "items": {
                        "type": "object",
                        "required": ["role", "content"],
                        "properties": {
                            "role": {
                                "type": "string",
                                "enum": ["system", "user", "assistant", "tool"],
                            },
                            "content": {
                                "description": 'List of Part objects (preferred) or a shorthand string. Preferred: [{"type":"text","text":"..."}].',
                                "anyOf": [
                                    {"type": "string"},
                                    {"type": "object"},
                                    {
                                        "type": "array",
                                        "minItems": 1,
                                        "examples": [
                                            [{"type": "text", "text": "Hello"}]
                                        ],
                                        "items": {
                                            "type": "object",
                                            "required": ["type"],
                                            "properties": {
                                                "type": {"type": "string"},
                                                "mime_type": {"type": "string"},
                                                "source": {
                                                    "anyOf": [
                                                        {
                                                            "type": "object",
                                                            "required": [
                                                                "kind",
                                                                "encoding",
                                                                "data",
                                                            ],
                                                            "properties": {
                                                                "kind": {
                                                                    "const": "bytes"
                                                                },
                                                                "encoding": {
                                                                    "const": "base64"
                                                                },
                                                                "data": {
                                                                    "type": "string"
                                                                },
                                                            },
                                                        },
                                                        {
                                                            "type": "object",
                                                            "required": ["kind", "url"],
                                                            "properties": {
                                                                "kind": {
                                                                    "const": "url"
                                                                },
                                                                "url": {
                                                                    "type": "string"
                                                                },
                                                            },
                                                        },
                                                        {
                                                            "type": "object",
                                                            "required": [
                                                                "kind",
                                                                "provider",
                                                                "id",
                                                            ],
                                                            "properties": {
                                                                "kind": {
                                                                    "const": "ref"
                                                                },
                                                                "provider": {
                                                                    "type": "string"
                                                                },
                                                                "id": {
                                                                    "type": "string"
                                                                },
                                                            },
                                                        },
                                                        {"type": "null"},
                                                    ]
                                                },
                                                "text": {"type": "string"},
                                                "embedding": {
                                                    "type": "array",
                                                    "items": {"type": "number"},
                                                },
                                                "meta": {"type": "object"},
                                            },
                                        },
                                    },
                                ],
                            },
                        },
                    },
                },
            ],
        },
        "output": {
            "type": "object",
            "required": ["modalities"],
            "properties": {
                "modalities": {"type": "array", "items": {"type": "string"}},
                "image": {
                    "type": "object",
                    "properties": {
                        "format": {
                            "type": "string",
                            "description": 'For image-only output, omit to default to "url" in MCP.',
                        }
                    },
                },
            },
        },
        "params": {"type": "object"},
        "wait": {"type": "boolean"},
        "tools": {
            "anyOf": [{"type": "array", "items": {"type": "object"}}, {"type": "null"}]
        },
        "tool_choice": {"anyOf": [{"type": "object"}, {"type": "null"}]},
        "provider_options": {"type": "object"},
    },
}


class ProvidersInfo(TypedDict):
    supported: list[str]
    configured: list[str]


class ModelInfo(TypedDict):
    model: str
    modes: list[str]
    input_modalities: list[str]
    output_modalities: list[str]


class AvailableModelsInfo(TypedDict):
    models: list[ModelInfo]


class McpGenerateResponseBase(TypedDict):
    id: str
    provider: str
    model: str
    status: str
    output: list[dict[str, Any]]


class McpGenerateResponse(McpGenerateResponseBase, total=False):
    usage: dict[str, Any] | None
    job: dict[str, Any] | None
    error: dict[str, Any] | None


_REQUEST_TOKEN: ContextVar[str | None] = ContextVar(
    "genai_calling_mcp_request_token", default=None
)

_MCP_BEARER_TOKEN_ENV = format_prefixed_env_names("MCP_BEARER_TOKEN")
_MCP_TOKEN_RULES_ENV = format_prefixed_env_names("MCP_TOKEN_RULES")


@dataclass(frozen=True, slots=True)
class McpTokenScope:
    providers_all: frozenset[str]
    models: dict[str, frozenset[str]]
    providers: frozenset[str]


_DENY_ALL_SCOPE = McpTokenScope(
    providers_all=frozenset(), models={}, providers=frozenset()
)


def _parse_mcp_token_scopes(raw: str) -> dict[str, McpTokenScope]:
    raw = raw.strip()
    if not raw:
        return {}

    # Accept a JSON dict: {"token": ["openai", "openai:gpt-4o-mini"], ...}
    if raw[0] == "{":
        try:
            parsed = json.loads(raw)
        except json.JSONDecodeError:
            parsed = None
        if isinstance(parsed, dict):
            items: list[str] = []
            for token, allow in parsed.items():
                if not isinstance(token, str) or not token.strip():
                    raise ValueError(
                        f"invalid {_MCP_TOKEN_RULES_ENV}: token must be a non-empty string"
                    )
                if allow is None:
                    items.append(f"{token.strip()}: []")
                    continue
                if not isinstance(allow, list) or not all(
                    isinstance(x, str) for x in allow
                ):
                    raise ValueError(
                        f"invalid {_MCP_TOKEN_RULES_ENV}: each token value must be a list of strings"
                    )
                joined = " ".join(a.strip() for a in allow if a.strip())
                items.append(f"{token.strip()}: [{joined}]")
            raw = ";".join(items)

    # Bracket syntax: token: [provider provider:model_id ...]; token2: [...]
    text = raw.replace("\\n", "\n")
    entries: list[str] = []
    for line in text.splitlines():
        for part in line.split(";"):
            stripped = part.strip()
            if stripped:
                entries.append(stripped)

    scopes: dict[str, McpTokenScope] = {}
    for entry in entries:
        if entry.startswith("#"):
            continue
        if ":" not in entry:
            raise ValueError(
                f"invalid {_MCP_TOKEN_RULES_ENV} entry (missing ':'): {entry}"
            )
        token, spec = entry.split(":", 1)
        token = token.strip()
        spec = spec.strip()
        if not token:
            raise ValueError(
                f"invalid {_MCP_TOKEN_RULES_ENV} entry (empty token): {entry}"
            )
        if token in scopes:
            raise ValueError(
                f"invalid {_MCP_TOKEN_RULES_ENV} (duplicate token): {token}"
            )
        if not (spec.startswith("[") and spec.endswith("]")):
            raise ValueError(
                f"invalid {_MCP_TOKEN_RULES_ENV} entry (expected '[...]'): {entry}"
            )
        inner = spec[1:-1].strip()
        providers_all: set[str] = set()
        models: dict[str, set[str]] = {}
        for item in inner.replace(",", " ").split():
            if not item:
                continue
            if ":" in item:
                provider, model_id = item.split(":", 1)
                provider = provider.strip().lower()
                model_id = model_id.strip()
                if not provider or not model_id:
                    raise ValueError(
                        f"invalid {_MCP_TOKEN_RULES_ENV} entry item: {item}"
                    )
                if model_id == "*":
                    providers_all.add(provider)
                    continue
                models.setdefault(provider, set()).add(model_id)
                continue
            providers_all.add(item.strip().lower())

        providers_all_fs = frozenset(providers_all)
        models_fs = {p: frozenset(ids) for p, ids in models.items() if ids}
        providers_fs = providers_all_fs | frozenset(models_fs.keys())
        scopes[token] = McpTokenScope(
            providers_all=providers_all_fs, models=models_fs, providers=providers_fs
        )

    return scopes


def _get_host_port() -> tuple[str, int]:
    host = (get_prefixed_env("MCP_HOST") or "").strip() or "127.0.0.1"
    port = get_prefixed_env_int("MCP_PORT", 6001)
    if port < 1:
        port = 1
    if port > 65535:
        port = 65535
    return host, port


def build_server(
    *,
    proxy_url: str | None = None,
    host: str | None = None,
    port: int | None = None,
    model_keywords: list[str] | None = None,
    bearer_token: str | None = None,
    token_scopes: dict[str, McpTokenScope] | None = None,
):
    """
    Build a FastMCP server that exposes:
    - generate: Client.generate wrapper (MCP-friendly defaults)
    - list_providers: discover providers configured on this server
    - list_available_models: list available models for a provider (fully-qualified)
    - list_all_available_models: list available models across all providers (fully-qualified)

    Notes for LLM tool callers:
    - Model must be "{provider}:{model_id}".
    - Call `list_providers` first if you don't know which providers are usable.
    """
    try:
        from mcp.server.fastmcp import FastMCP
    except ModuleNotFoundError as e:  # pragma: no cover
        raise SystemExit(
            "missing dependency: install `mcp` to run the MCP server (e.g. `uv sync`)"
        ) from e

    try:
        from pydantic import Field, WithJsonSchema
    except ModuleNotFoundError as e:  # pragma: no cover
        raise SystemExit(
            "missing dependency: install `mcp` to run the MCP server (e.g. `uv sync`)"
        ) from e
    from typing import Annotated
    from starlette.requests import Request
    from starlette.responses import Response

    os.environ["GENAI_CALLING_TRANSPORT"] = "mcp"
    if host is None or port is None:
        host, port = _get_host_port()
    server = FastMCP(name="genai-calling", host=host, port=port)

    keywords: list[str] = []
    for raw in model_keywords or []:
        if not isinstance(raw, str):
            continue
        for part in raw.split(","):
            keyword = part.strip().lower()
            if keyword:
                keywords.append(keyword)

    def _model_allowed(model: str) -> bool:
        if not keywords:
            return True
        candidate = model.strip().lower()
        return any(k in candidate for k in keywords)

    max_artifacts = get_prefixed_env_int("MAX_ARTIFACTS", 64)
    if max_artifacts < 1:
        max_artifacts = 1
    max_artifact_bytes = get_prefixed_env_int("MAX_ARTIFACT_BYTES", 64 * 1024 * 1024)
    if max_artifact_bytes < 0:
        max_artifact_bytes = 0
    artifact_url_ttl_seconds = get_prefixed_env_int("ARTIFACT_URL_TTL_SECONDS", 600)
    if artifact_url_ttl_seconds < 1:
        artifact_url_ttl_seconds = 1

    def _public_base_url() -> str:
        base = (get_prefixed_env("MCP_PUBLIC_BASE_URL") or "").strip()
        if base:
            return base.rstrip("/")
        if host in {"0.0.0.0", "::"}:
            return f"http://127.0.0.1:{port}"
        return f"http://{host}:{port}"

    base_url = _public_base_url()
    signing_token = (
        bearer_token.strip()
        if isinstance(bearer_token, str) and bearer_token.strip()
        else None
    )
    token_scopes = token_scopes or {}

    def _b64url(data: bytes) -> str:
        return base64.urlsafe_b64encode(data).decode("ascii").rstrip("=")

    def _signing_key() -> str | None:
        current = _REQUEST_TOKEN.get()
        if isinstance(current, str) and current:
            return current
        return signing_token

    def _artifact_sig(key: str, artifact_id: str, exp: int) -> str:
        msg = f"{artifact_id}.{exp}".encode("utf-8", errors="strict")
        digest = hmac.new(
            key.encode("utf-8", errors="strict"), msg, hashlib.sha256
        ).digest()
        return _b64url(digest)

    def _artifact_url(artifact_id: str) -> str:
        url = f"{base_url}/artifact/{artifact_id}"
        key = _signing_key()
        if key is None:
            return url
        exp = int(time.time()) + artifact_url_ttl_seconds
        sig = _artifact_sig(key, artifact_id, exp)
        return f"{url}?exp={exp}&sig={sig}"

    @dataclass(slots=True)
    class _ArtifactItem:
        data: bytes
        mime_type: str | None
        owner_token: str | None

    artifacts: OrderedDict[str, _ArtifactItem] = OrderedDict()
    artifacts_total_bytes = 0

    def _evict_one() -> None:
        nonlocal artifacts_total_bytes
        _, item = artifacts.popitem(last=False)
        artifacts_total_bytes -= len(item.data)

    def _enforce_artifact_limits() -> None:
        while len(artifacts) > max_artifacts:
            _evict_one()
        if max_artifact_bytes <= 0:
            while artifacts:
                _evict_one()
            return
        while artifacts and artifacts_total_bytes > max_artifact_bytes:
            _evict_one()

    def _artifact_owner() -> str | None:
        token = _REQUEST_TOKEN.get()
        if isinstance(token, str) and token:
            return token
        return signing_token

    def _store_artifact(data: bytes, mime_type: str | None) -> str | None:
        nonlocal artifacts_total_bytes
        if max_artifact_bytes <= 0:
            return None
        if len(data) > max_artifact_bytes:
            return None
        artifact_id = uuid4().hex
        artifacts[artifact_id] = _ArtifactItem(
            data=data, mime_type=mime_type, owner_token=_artifact_owner()
        )
        artifacts_total_bytes += len(data)
        artifacts.move_to_end(artifact_id)
        _enforce_artifact_limits()
        return artifact_id if artifact_id in artifacts else None

    class _McpArtifactStore:
        def put(self, data: bytes, mime_type: str | None) -> str | None:
            return _store_artifact(data, mime_type)

        def url(self, artifact_id: str) -> str:
            return _artifact_url(artifact_id)

    client = Client(
        proxy_url=proxy_url,
        artifact_store=_McpArtifactStore(),
    )

    @server.custom_route(
        "/artifact/{artifact_id}", methods=["GET", "HEAD"], include_in_schema=False
    )
    async def artifact_route(request: Request) -> Response:
        artifact_id = request.path_params.get("artifact_id")
        if not isinstance(artifact_id, str) or not artifact_id:
            return Response(status_code=404)
        item = artifacts.get(artifact_id)
        if item is None:
            return Response(status_code=404)
        owner = item.owner_token
        if owner is not None:
            token = _REQUEST_TOKEN.get()
            if token is None or not compare_digest(token, owner):
                return Response(status_code=404)
        artifacts.move_to_end(artifact_id)
        if item.mime_type is None:
            guessed = sniff_image_mime_type(item.data)
            if guessed is not None:
                item.mime_type = guessed
        headers = {"Content-Length": str(len(item.data))}
        if request.method.upper() == "HEAD":
            return Response(
                content=b"",
                media_type=item.mime_type or "application/octet-stream",
                headers=headers,
            )
        return Response(
            content=item.data,
            media_type=item.mime_type or "application/octet-stream",
            headers=headers,
        )

    @server.resource("genai://artifact/{artifact_id}", mime_type="application/json")
    def read_artifact(artifact_id: str) -> dict[str, Any]:
        item = artifacts.get(artifact_id)
        if item is None:
            raise ValueError("artifact not found")
        owner = item.owner_token
        if owner is not None:
            token = _REQUEST_TOKEN.get()
            if token is None or not compare_digest(token, owner):
                raise ValueError("artifact not found")
        artifacts.move_to_end(artifact_id)
        if item.mime_type is None:
            item.mime_type = sniff_image_mime_type(item.data)
        return {
            "id": artifact_id,
            "mime_type": item.mime_type,
            "bytes": len(item.data),
            "url": _artifact_url(artifact_id),
        }

    def _scope() -> McpTokenScope | None:
        if not token_scopes:
            return None
        token = _REQUEST_TOKEN.get()
        if not token:
            return None
        return token_scopes.get(token, _DENY_ALL_SCOPE)

    def _provider_allowed(provider: str) -> bool:
        scope = _scope()
        if scope is None:
            return True
        return provider in scope.providers

    def _model_allowed_by_scope(model: str) -> bool:
        scope = _scope()
        if scope is None:
            return True
        if ":" not in model:
            return False
        provider, model_id = model.split(":", 1)
        provider = provider.strip().lower()
        model_id = model_id.strip()
        if not provider or not model_id:
            return False
        if provider in scope.providers_all:
            return True
        allowed = scope.models.get(provider)
        if allowed is None:
            return False
        return model_id in allowed

    def list_providers() -> ProvidersInfo:
        """
        List providers supported by this MCP server.

        Returns:
        - supported: providers known by the SDK catalog
        - configured: providers that have credentials configured and can be used
        """
        from .reference import get_model_catalog
        from ._internal.errors import GenAIError

        supported = sorted(get_model_catalog().keys())
        if token_scopes:
            supported = [p for p in supported if _provider_allowed(p)]
        configured: list[str] = []
        for p in supported:
            try:
                client._adapter(p)  # type: ignore[attr-defined]
            except GenAIError:
                continue
            configured.append(p)
        return {"supported": supported, "configured": configured}

    def list_available_models(
        provider: str, *, timeout_ms: int | None = None
    ) -> AvailableModelsInfo:
        """
        List available models (sdk catalog ∩ remotely available) for a provider.

        Always returns fully-qualified model strings like "openai:gpt-4o-mini".

        Returns:
        - models: list[object], each includes:
          - model: "{provider}:{model_id}"
          - modes: ["sync","stream","job","async"] (varies by provider/model)
          - input_modalities: ["text","image",...]
          - output_modalities: ["text","image",...]
        """
        from .client import _normalize_provider
        from .reference import get_sdk_supported_models_for_provider

        p = _normalize_provider(provider)
        if not _provider_allowed(p):
            raise ValueError(f"provider not allowed: {p}")
        ids = client.list_available_models(p, timeout_ms=timeout_ms)
        rows = get_sdk_supported_models_for_provider(p)
        by_model_id = {r["model_id"]: r for r in rows}

        models: list[ModelInfo] = []
        for model_id in ids:
            row = by_model_id.get(model_id)
            if row is None:
                continue
            model = row["model"]
            if not _model_allowed(model) or not _model_allowed_by_scope(model):
                continue
            models.append(
                {
                    "model": model,
                    "modes": row["modes"],
                    "input_modalities": row["input_modalities"],
                    "output_modalities": row["output_modalities"],
                }
            )
        return {"models": models}

    def list_all_available_models(
        *, timeout_ms: int | None = None
    ) -> AvailableModelsInfo:
        """
        List available models (sdk catalog ∩ remotely available) across all providers.

        Always returns fully-qualified model strings like "openai:gpt-4o-mini".

        Returns:
        - models: list[object], each includes:
          - model: "{provider}:{model_id}"
          - modes: ["sync","stream","job","async"] (varies by provider/model)
          - input_modalities: ["text","image",...]
          - output_modalities: ["text","image",...]
        """
        scope = _scope()
        if scope is None:
            from .reference import get_sdk_supported_models

            rows = get_sdk_supported_models()
            by_model = {r["model"]: r for r in rows}

            out_models: list[ModelInfo] = []
            for model in client.list_all_available_models(timeout_ms=timeout_ms):
                if not _model_allowed(model) or not _model_allowed_by_scope(model):
                    continue
                row = by_model.get(model)
                if row is None:
                    continue
                out_models.append(
                    {
                        "model": model,
                        "modes": row["modes"],
                        "input_modalities": row["input_modalities"],
                        "output_modalities": row["output_modalities"],
                    }
                )
            return {"models": out_models}

        models: list[ModelInfo] = []
        for provider in sorted(scope.providers):
            models.extend(
                list_available_models(provider, timeout_ms=timeout_ms)["models"]
            )
        models.sort(key=lambda row: row["model"])
        return {"models": models}

    def generate(
        request: dict[str, Any], *, stream: bool = False
    ) -> McpGenerateResponse:
        """
        MCP-friendly wrapper of `Client.generate`.

        Behavior:
        - Enforces non-stream tool behavior (MCP tool call returns one result).
        - For image-only output, defaults to URL output when format is unspecified.
        - Large base64 binary outputs are stored on this server and returned as a URL.

        Notes:
        - `request.model` must be "{provider}:{model_id}" (e.g. "openai:gpt-4o-mini").
        - `request.input` preferred format: [{"role":"user","content":[{"type":"text","text":"..."}]}].
          Shorthand accepted: `request.input="..."` or [{"role":"user","content":"..."}] (auto-coerced to a text Part).
        - Call `list_providers` / `list_available_models` / `list_all_available_models` first when in doubt.

        Returns:
        - id/provider/model/status/output (+ optional usage/job/error)
        """
        if stream:
            raise ValueError(
                "stream=true is not supported for MCP tool calls; use stream=false"
            )

        from pydantic import TypeAdapter, ValidationError

        if not isinstance(request, dict):
            raise ValueError("request must be an object")

        req_dict: dict[str, Any] = dict(request)
        msgs = req_dict.get("input")
        if isinstance(msgs, str):
            msgs = [{"role": "user", "content": msgs}]
        elif isinstance(msgs, dict):
            msgs = [msgs]
        if isinstance(msgs, list):
            coerced_msgs: list[Any] = []
            for msg in msgs:
                if isinstance(msg, str):
                    coerced_msgs.append({"role": "user", "content": msg})
                    continue
                if not isinstance(msg, dict):
                    coerced_msgs.append(msg)
                    continue
                m = dict(msg)
                if not isinstance(m.get("role"), str):
                    m["role"] = "user"
                content = m.get("content")
                if isinstance(content, str):
                    m["content"] = [{"type": "text", "text": content}]
                elif isinstance(content, dict):
                    part = dict(content)
                    if "type" not in part and isinstance(part.get("text"), str):
                        part["type"] = "text"
                    m["content"] = [part]
                elif isinstance(content, list):
                    parts: list[Any] = []
                    for part in content:
                        if isinstance(part, str):
                            parts.append({"type": "text", "text": part})
                        elif (
                            isinstance(part, dict)
                            and "type" not in part
                            and isinstance(part.get("text"), str)
                        ):
                            p = dict(part)
                            p["type"] = "text"
                            parts.append(p)
                        else:
                            parts.append(part)
                    m["content"] = parts
                coerced_msgs.append(m)
            req_dict["input"] = coerced_msgs

        out_spec = req_dict.get("output")
        if isinstance(out_spec, str):
            req_dict["output"] = {"modalities": [out_spec]}
        elif isinstance(out_spec, dict) and isinstance(out_spec.get("modalities"), str):
            o = dict(out_spec)
            o["modalities"] = [out_spec["modalities"]]
            req_dict["output"] = o

        try:
            req = TypeAdapter(GenerateRequest).validate_python(req_dict)
        except ValidationError as e:
            raise ValueError(str(e)) from e

        if not _model_allowed(req.model):
            raise ValueError(f"model not allowed by server filter: {req.model}")
        if not _model_allowed_by_scope(req.model):
            raise ValueError(f"model not allowed: {req.model}")

        if set(req.output.modalities) == {"image"}:
            img = req.output.image or OutputImageSpec()
            if not img.format:
                img = replace(img, format="url")
                req = replace(req, output=replace(req.output, image=img))

        resp = client.generate(req, stream=False)
        if not isinstance(resp, GenerateResponse):
            raise ValueError(
                "provider returned stream response; expected non-stream response"
            )
        return cast(McpGenerateResponse, asdict(resp))

    generate.__annotations__["request"] = Annotated[
        dict[str, Any], WithJsonSchema(_MCP_GENERATE_REQUEST_SCHEMA)
    ]
    list_available_models.__annotations__["provider"] = Annotated[
        str,
        Field(
            description='Provider name (e.g. "openai"). Call `list_providers` first if unknown.',
            examples=["openai"],
        ),
    ]
    list_available_models.__annotations__["timeout_ms"] = Annotated[
        int | None,
        Field(
            default=None,
            description="Remote request timeout in milliseconds.",
        ),
    ]
    list_all_available_models.__annotations__["timeout_ms"] = Annotated[
        int | None,
        Field(
            default=None,
            description="Remote request timeout in milliseconds.",
        ),
    ]

    server.tool(structured_output=True)(generate)
    server.tool(structured_output=True)(list_providers)
    server.tool(structured_output=True)(list_available_models)
    server.tool(structured_output=True)(list_all_available_models)
    return server


def build_http_app(server: Any) -> Any:
    from starlette.routing import Mount, Route

    app = server.streamable_http_app()
    sse = server.sse_app()
    sse_path = server.settings.sse_path
    message_path = server.settings.message_path.rstrip("/")

    for route in sse.router.routes:
        if isinstance(route, Route) and route.path == sse_path:
            app.router.routes.append(route)
            continue
        if isinstance(route, Mount) and route.path.rstrip("/") == message_path:
            app.router.routes.append(route)
            continue
    return app


def main(argv: list[str] | None = None) -> None:
    from ._internal.config import load_env_files

    load_env_files()

    parser = argparse.ArgumentParser(
        prog="genai-mcp-server",
        description="genai-calling MCP server (Streamable HTTP: /mcp, SSE: /sse)",
    )
    parser.add_argument(
        "--proxy",
        dest="proxy_url",
        help="HTTP proxy URL for provider requests (e.g. http://127.0.0.1:7890)",
    )
    parser.add_argument(
        "--bearer-token",
        dest="bearer_token",
        help="Require HTTP Authorization: Bearer <token> for all endpoints (or set GENAI_CALLING_MCP_BEARER_TOKEN).",
    )
    parser.add_argument(
        "--model-keyword",
        dest="model_keywords",
        action="append",
        help='Only expose models whose "{provider}:{model_id}" contains this substring (case-insensitive). Repeatable; comma-separated also accepted.',
    )
    args = parser.parse_args(argv)
    bearer = (args.bearer_token or get_prefixed_env("MCP_BEARER_TOKEN") or "").strip()
    token_rules = (get_prefixed_env("MCP_TOKEN_RULES") or "").strip()
    if token_rules and bearer:
        raise SystemExit(
            f"set either {_MCP_BEARER_TOKEN_ENV}/--bearer-token or {_MCP_TOKEN_RULES_ENV}, not both"
        )
    token_scopes: dict[str, McpTokenScope] = {}
    if token_rules:
        try:
            token_scopes = _parse_mcp_token_scopes(token_rules)
        except ValueError as e:
            raise SystemExit(str(e)) from e
        if not token_scopes:
            raise SystemExit(f"invalid {_MCP_TOKEN_RULES_ENV}: no tokens configured")

    server_host, server_port = _get_host_port()
    server = build_server(
        proxy_url=args.proxy_url,
        host=server_host,
        port=server_port,
        model_keywords=args.model_keywords,
        bearer_token=bearer,
        token_scopes=token_scopes,
    )
    app = build_http_app(server)
    if token_scopes:
        app.add_middleware(_BearerAuthMiddleware, tokens=token_scopes)
    elif bearer:
        app.add_middleware(_BearerAuthMiddleware, token=bearer)

    try:
        import uvicorn
    except ModuleNotFoundError as e:  # pragma: no cover
        raise SystemExit(
            "missing dependency: install `uvicorn` to run the MCP server (e.g. `uv sync`)"
        ) from e

    uvicorn.run(
        app,
        host=server_host,
        port=server_port,
        log_level=server.settings.log_level.lower(),
    )


class _BearerAuthMiddleware:
    def __init__(
        self,
        app: Any,
        *,
        token: str | None = None,
        tokens: dict[str, McpTokenScope] | None = None,
    ) -> None:
        self.app = app
        token = token.strip() if isinstance(token, str) else ""
        tokens = tokens or {}
        if token and tokens:
            raise ValueError("pass either token=... or tokens=..., not both")
        if not token and not tokens:
            raise ValueError("missing auth config: pass token=... or tokens=...")
        self._tokens = [token] if token else list(tokens.keys())

    def _match_bearer(self, scope: Any) -> str | None:
        raw = None
        for k, v in scope.get("headers") or []:
            if k.lower() == b"authorization":
                raw = v
                break
        if not raw:
            return None

        try:
            header = raw.decode("utf-8", errors="replace").strip()
        except Exception:
            return None

        if not header.lower().startswith("bearer "):
            return None

        token = header[7:].strip()
        for candidate in self._tokens:
            if compare_digest(token, candidate):
                return candidate
        return None

    def _match_artifact_sig(self, scope: Any) -> str | None:
        if not self._tokens:
            return None

        path = scope.get("path")
        if not isinstance(path, str) or not path.startswith("/artifact/"):
            return None

        method = (scope.get("method") or "").upper()
        if method not in {"GET", "HEAD"}:
            return None

        artifact_id = path[len("/artifact/") :].strip("/")
        if not artifact_id:
            return None

        qs = scope.get("query_string") or b""
        try:
            query = parse_qs(
                qs.decode("utf-8", errors="replace"), keep_blank_values=True
            )
        except Exception:
            return None

        exp_raw = (query.get("exp") or [None])[0]
        sig_raw = (query.get("sig") or [None])[0]
        if (
            not isinstance(exp_raw, str)
            or not exp_raw.strip()
            or not isinstance(sig_raw, str)
            or not sig_raw.strip()
        ):
            return None

        try:
            exp = int(exp_raw.strip())
        except ValueError:
            return None

        if int(time.time()) > exp:
            return None

        msg = f"{artifact_id}.{exp}".encode("utf-8", errors="strict")
        sig = sig_raw.strip()
        for token in self._tokens:
            digest = hmac.new(
                token.encode("utf-8", errors="strict"), msg, hashlib.sha256
            ).digest()
            expected = base64.urlsafe_b64encode(digest).decode("ascii").rstrip("=")
            if compare_digest(sig, expected):
                return token
        return None

    async def __call__(self, scope: Any, receive: Any, send: Any) -> None:
        if scope.get("type") != "http":
            await self.app(scope, receive, send)
            return

        token = self._match_bearer(scope) or self._match_artifact_sig(scope)
        if token is None:
            await _send_unauthorized(send)
            return

        ctx = _REQUEST_TOKEN.set(token)
        try:
            await self.app(scope, receive, send)
        finally:
            _REQUEST_TOKEN.reset(ctx)


async def _send_unauthorized(send: Any) -> None:
    body = b'{"error":"invalid_token","error_description":"Authentication required"}'
    await send(
        {
            "type": "http.response.start",
            "status": 401,
            "headers": [
                (b"content-type", b"application/json"),
                (b"content-length", str(len(body)).encode()),
                (
                    b"www-authenticate",
                    b'Bearer error="invalid_token", error_description="Authentication required"',
                ),
            ],
        }
    )
    await send({"type": "http.response.body", "body": body})


if __name__ == "__main__":  # pragma: no cover
    main()



"""
Central place for capability rules based on model series (usually shared prefixes).

Rules here only look at model_id strings and are provider-agnostic. Provider adapters
still own protocol details (stream/job semantics, routing, endpoint quirks).
"""

from __future__ import annotations

import re
from typing import Final, Literal

from ..types import Modality

ModelKind = Literal["video", "image", "embedding", "tts", "transcribe", "chat"]
GeminiModelKind = Literal["video", "embedding", "tts", "native_audio", "image", "chat"]


def _mods(*values: Modality) -> set[Modality]:
    return set(values)


def _norm(model_id: str) -> str:
    return model_id.lower().strip()


def _starts_with_any(s: str, prefixes: tuple[str, ...]) -> bool:
    return s.startswith(prefixes)


_TOKEN_SPLIT_RE: Final[re.Pattern[str]] = re.compile(r"[^a-z0-9]+")


def _has_token(s: str, token: str) -> bool:
    # Token boundary: split by any non [a-z0-9] to avoid substring false-positives.
    toks = _TOKEN_SPLIT_RE.split(s.lower())
    return token.lower() in toks


# ---- Model series for /v1-style protocols ----

_SORA_PREFIX: Final[str] = "sora-"

_VEO_PREFIX: Final[str] = "veo"
_PIKA_PREFIX: Final[str] = "pika-"
_RUNWAY_PREFIX: Final[str] = "runway-"
_SEEDANCE_PREFIXES: Final[tuple[str, ...]] = ("seedance", "doubao-seedance")
_KLING_VIDEO_PREFIXES: Final[tuple[str, ...]] = ("kling_video", "kling-video-")

_DALLE_PREFIX: Final[str] = "dall-e-"
_GPT_IMAGE_PREFIX: Final[str] = "gpt-image-"
_CHATGPT_IMAGE_PREFIX: Final[str] = "chatgpt-image"
_SEEDREAM_PREFIXES: Final[tuple[str, ...]] = ("seedream", "doubao-seedream")
_SEEDEDIT_PREFIXES: Final[tuple[str, ...]] = ("seededit", "api-images-seededit")
_FLUX_PREFIXES: Final[tuple[str, ...]] = ("flux-", "flux.")
_SD3_PREFIX: Final[str] = "sd3"

_TEXT_EMBEDDING_PREFIX: Final[str] = "text-embedding-"
_EMBEDDING_PREFIX: Final[str] = "embedding-"
_DOUBAO_EMBEDDING_PREFIX: Final[str] = "doubao-embedding-"

_TTS_PREFIX: Final[str] = "tts-"
_TTS_SUFFIX: Final[str] = "-tts"
_VOICE_SUFFIXES: Final[tuple[str, ...]] = ("-voice", "_voice")
_ADVANCED_VOICE_MODEL: Final[str] = "advanced-voice"
_CHIRP_PREFIX: Final[str] = "chirp-"

_WHISPER_PREFIX: Final[str] = "whisper-"
_DISTIL_WHISPER_PREFIX: Final[str] = "distil-whisper-"
_TRANSCRIBE_MARKER: Final[str] = "-transcribe"
_ASR_PREFIX: Final[str] = "asr"

_IMAGE_MODELS_WITH_IMAGE_INPUT_PREFIXES: Final[tuple[str, ...]] = (
    "gpt-image-1",
    "chatgpt-image",
)

_Z_IMAGE_PREFIX: Final[str] = "z-image"

_CHAT_IMAGE_INPUT_EXACT: Final[tuple[str, ...]] = (
    "chatgpt-4o-latest",
    "codex-mini-latest",
    "computer-use-preview",
    "omni-moderation-latest",
)

_DEEPSEEK_V3_PREFIX: Final[str] = "deepseek-v3"
_KIMI_K2_PREFIX: Final[str] = "kimi-k2"
_KIMI_LATEST_PREFIX: Final[str] = "kimi-latest"
_QWEN_PREFIX: Final[str] = "qwen"
_QVQ_PREFIX: Final[str] = "qvq"

_CHAT_NO_IMAGE_PREFIXES: Final[tuple[str, ...]] = (
    _DEEPSEEK_V3_PREFIX,
    _KIMI_K2_PREFIX,
    _QVQ_PREFIX,
)

_CHAT_TEXT_ONLY_EXACT: Final[tuple[str, ...]] = (
    "gpt-4",
    "gpt-4-turbo-preview",
    "o1-mini",
    "o1-preview",
    "o3-mini",
)

_CHAT_TEXT_ONLY_MARKERS: Final[tuple[str, ...]] = ("search-preview",)

_CHAT_AUDIO_PREFIXES: Final[tuple[str, ...]] = ("gpt-audio", "gpt-realtime")
_CHAT_AUDIO_SUFFIXES: Final[tuple[str, ...]] = ("-audio-preview", "-realtime-preview")

_CLAUDE_PREFIX: Final[str] = "claude-"


def is_sora_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_SORA_PREFIX)


def is_veo_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_VEO_PREFIX)


def is_pika_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_PIKA_PREFIX)


def is_runway_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_RUNWAY_PREFIX)


def is_seedance_model(model_id: str) -> bool:
    return _starts_with_any(_norm(model_id), _SEEDANCE_PREFIXES)


def is_kling_video_model(model_id: str) -> bool:
    return _starts_with_any(_norm(model_id), _KLING_VIDEO_PREFIXES)


def is_video_model(model_id: str) -> bool:
    return (
        is_sora_model(model_id)
        or is_veo_model(model_id)
        or is_pika_model(model_id)
        or is_runway_model(model_id)
        or is_seedance_model(model_id)
        or is_kling_video_model(model_id)
    )


def is_dall_e_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_DALLE_PREFIX)


def is_gpt_image_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_GPT_IMAGE_PREFIX)


def is_chatgpt_image_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_CHATGPT_IMAGE_PREFIX)


def is_seedream_model(model_id: str) -> bool:
    return _starts_with_any(_norm(model_id), _SEEDREAM_PREFIXES)


def is_seededit_model(model_id: str) -> bool:
    return _starts_with_any(_norm(model_id), _SEEDEDIT_PREFIXES)


def is_flux_model(model_id: str) -> bool:
    return _starts_with_any(_norm(model_id), _FLUX_PREFIXES)


def is_sd3_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_SD3_PREFIX)


def is_z_image_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_Z_IMAGE_PREFIX)


def is_gpt_dash_image_model(model_id: str) -> bool:
    mid_l = _norm(model_id)
    return mid_l.startswith("gpt-") and "-image" in mid_l


def is_image_model(model_id: str) -> bool:
    return (
        is_dall_e_model(model_id)
        or is_gpt_image_model(model_id)
        or is_gpt_dash_image_model(model_id)
        or is_chatgpt_image_model(model_id)
        or is_z_image_model(model_id)
        or is_seedream_model(model_id)
        or is_seededit_model(model_id)
        or is_flux_model(model_id)
        or is_sd3_model(model_id)
    )


def is_text_embedding_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_TEXT_EMBEDDING_PREFIX)


def is_embedding_series_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_EMBEDDING_PREFIX)


def is_doubao_embedding_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_DOUBAO_EMBEDDING_PREFIX)


def is_embedding_model(model_id: str) -> bool:
    return (
        is_text_embedding_model(model_id)
        or is_embedding_series_model(model_id)
        or is_doubao_embedding_model(model_id)
    )


def is_tts_model(model_id: str) -> bool:
    mid_l = _norm(model_id)
    return (
        mid_l.startswith(_TTS_PREFIX)
        or mid_l.startswith(_CHIRP_PREFIX)
        or mid_l.endswith(_TTS_SUFFIX)
        or mid_l.endswith(_VOICE_SUFFIXES)
        or mid_l == _ADVANCED_VOICE_MODEL
    )


def is_whisper_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_WHISPER_PREFIX)


def is_distil_whisper_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_DISTIL_WHISPER_PREFIX)


def is_transcribe_marker_model(model_id: str) -> bool:
    return _TRANSCRIBE_MARKER in _norm(model_id)


def is_asr_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_ASR_PREFIX)


def is_transcribe_model(model_id: str) -> bool:
    return (
        is_whisper_model(model_id)
        or is_distil_whisper_model(model_id)
        or is_transcribe_marker_model(model_id)
        or is_asr_model(model_id)
    )


def infer_model_kind(model_id: str) -> ModelKind:
    if is_video_model(model_id):
        return "video"
    if is_image_model(model_id):
        return "image"
    if is_embedding_model(model_id):
        return "embedding"
    if is_tts_model(model_id):
        return "tts"
    if is_transcribe_model(model_id):
        return "transcribe"
    return "chat"


def output_modalities_for_kind(kind: ModelKind) -> set[Modality] | None:
    if kind == "video":
        return _mods("video")
    if kind == "image":
        return _mods("image")
    if kind == "embedding":
        return _mods("embedding")
    if kind == "tts":
        return _mods("audio")
    if kind == "transcribe":
        return _mods("text")
    return None


def transcribe_input_modalities(_: str) -> set[Modality]:
    return _mods("audio")


def is_claude_model(model_id: str) -> bool:
    return _norm(model_id).startswith(_CLAUDE_PREFIX)


def claude_input_modalities(model_id: str) -> set[Modality]:
    if is_claude_model(model_id):
        return _mods("text", "image")
    return _mods("text")


def image_input_modalities(model_id: str) -> set[Modality]:
    mid_l = _norm(model_id)
    if _starts_with_any(
        mid_l, _IMAGE_MODELS_WITH_IMAGE_INPUT_PREFIXES
    ) or mid_l.startswith(_Z_IMAGE_PREFIX):
        return _mods("text", "image")
    return _mods("text")


def video_input_modalities(model_id: str) -> set[Modality]:
    if is_veo_model(model_id):
        return _mods("text", "image")
    return _mods("text")


def chat_input_modalities(model_id: str) -> set[Modality]:
    mid_l = _norm(model_id)
    out: set[Modality] = {"text"}
    if _chat_supports_image_input(mid_l):
        out.add("image")
    if _chat_supports_audio_input(mid_l):
        out.add("audio")
    return out


def chat_output_modalities(model_id: str) -> set[Modality]:
    mid_l = _norm(model_id)
    out: set[Modality] = {"text"}
    if _chat_supports_audio_io(mid_l):
        out.add("audio")
    return out


def chat_supports_audio_io(model_id: str) -> bool:
    return _chat_supports_audio_io(_norm(model_id))


def _chat_supports_audio_io(mid_l: str) -> bool:
    return _starts_with_any(mid_l, _CHAT_AUDIO_PREFIXES) or mid_l.endswith(
        _CHAT_AUDIO_SUFFIXES
    )


def _qwen_supports_image_input(mid_l: str) -> bool:
    return _has_token(mid_l, "image") or _has_token(mid_l, "vl")


def _qwen_supports_audio_input(mid_l: str) -> bool:
    return _has_token(mid_l, "asr")


def chat_supports_audio_input(model_id: str) -> bool:
    return _chat_supports_audio_input(_norm(model_id))


def _chat_supports_audio_input(mid_l: str) -> bool:
    if _chat_supports_audio_io(mid_l):
        return True
    return mid_l.startswith(_QWEN_PREFIX) and _qwen_supports_audio_input(mid_l)


def chat_supports_image_input(model_id: str) -> bool:
    return _chat_supports_image_input(_norm(model_id))


def _chat_supports_image_input(mid_l: str) -> bool:
    if mid_l in _CHAT_TEXT_ONLY_EXACT:
        return False
    if any(marker in mid_l for marker in _CHAT_TEXT_ONLY_MARKERS):
        return False

    if _starts_with_any(mid_l, _CHAT_NO_IMAGE_PREFIXES):
        return False
    if mid_l.startswith(_KIMI_LATEST_PREFIX):
        return True
    if mid_l.startswith(_QWEN_PREFIX):
        return _qwen_supports_image_input(mid_l)

    if mid_l in _CHAT_IMAGE_INPUT_EXACT:
        return True

    return _openai_style_chat_supports_image_input(mid_l)


def _openai_style_chat_supports_image_input(mid_l: str) -> bool:
    if mid_l.startswith("gpt-realtime"):
        return True
    if mid_l.startswith("gpt-4o") and not _chat_supports_audio_io(mid_l):
        return True
    if mid_l.startswith("gpt-4-turbo") and mid_l != "gpt-4-turbo-preview":
        return True
    if mid_l.startswith(("gpt-4.1", "gpt-4.5", "gpt-5")):
        return True
    if mid_l == "o1" or mid_l.startswith("o1-"):
        return True
    if mid_l == "o3" or mid_l.startswith("o3-"):
        return True
    return mid_l.startswith("o4-")


# ---- Gemini model series ----


def _gemini_norm(model_id: str) -> str:
    mid = model_id.lower().strip()
    if mid.startswith("models/"):
        mid = mid[len("models/") :]
    return mid


def _gemini_is_image_model(mid_l: str) -> bool:
    return (
        mid_l.startswith("imagen-")
        or "image-generation" in mid_l
        or mid_l.endswith("-image")
        or mid_l.endswith("-image-preview")
        or "-image-" in mid_l
    )


def gemini_is_video_model(model_id: str) -> bool:
    return _gemini_norm(model_id).startswith("veo")


def gemini_is_embedding_model(model_id: str) -> bool:
    return "embedding" in _gemini_norm(model_id)


def gemini_is_tts_model(model_id: str) -> bool:
    return "tts" in _gemini_norm(model_id)


def gemini_is_native_audio_model(model_id: str) -> bool:
    return "native-audio" in _gemini_norm(model_id)


def gemini_is_image_model(model_id: str) -> bool:
    return _gemini_is_image_model(_gemini_norm(model_id))


def gemini_image_input_modalities(model_id: str) -> set[Modality]:
    mid_l = _gemini_norm(model_id)
    if mid_l.startswith("imagen-"):
        return _mods("text")
    if "image-generation" in mid_l:
        return _mods("text")
    return _mods("text", "image")


def gemini_model_kind(model_id: str) -> GeminiModelKind:
    mid_l = _gemini_norm(model_id)
    if mid_l.startswith("veo"):
        return "video"
    if "embedding" in mid_l:
        return "embedding"
    if "tts" in mid_l:
        return "tts"
    if "native-audio" in mid_l:
        return "native_audio"
    if _gemini_is_image_model(mid_l):
        return "image"
    return "chat"


def gemini_output_modalities(kind: GeminiModelKind) -> set[Modality]:
    if kind == "video":
        return _mods("video")
    if kind == "embedding":
        return _mods("embedding")
    if kind == "tts":
        return _mods("audio")
    if kind == "native_audio":
        return _mods("text", "audio")
    if kind == "image":
        return _mods("image")
    return _mods("text")



from __future__ import annotations

MODELS: list[str] = [
    "deepseek-r1-250120",
    "deepseek-r1-250528",
    "deepseek-r1-distill-qwen-32b-250120",
    "deepseek-r1-distill-qwen-7b-250120",
    "deepseek-v3-1-250821",
    "deepseek-v3-1-terminus",
    "deepseek-v3-2-251201",
    "deepseek-v3-241226",
    "deepseek-v3-250324",
    "doubao-1-5-lite-32k-250115",
    "doubao-1-5-pro-256k-250115",
    "doubao-1-5-pro-32k-250115",
    "doubao-1-5-pro-32k-character-250228",
    "doubao-1-5-pro-32k-character-250715",
    "doubao-1-5-thinking-pro-250415",
    "doubao-1-5-thinking-pro-m-250415",
    "doubao-1-5-thinking-pro-m-250428",
    "doubao-1-5-thinking-vision-pro-250428",
    "doubao-1-5-ui-tars-250428",
    "doubao-1-5-vision-pro-32k-250115",
    "doubao-1.5-ui-tars-250328",
    "doubao-1.5-vision-lite-250315",
    "doubao-1.5-vision-pro-250328",
    "doubao-embedding-vision-241215",
    "doubao-embedding-vision-250328",
    "doubao-embedding-vision-250615",
    "doubao-embedding-vision-251215",
    "doubao-lite-128k-240428",
    "doubao-lite-128k-240828",
    "doubao-lite-32k-240428",
    "doubao-lite-32k-240628",
    "doubao-lite-32k-240828",
    "doubao-lite-32k-character-241015",
    "doubao-lite-32k-character-250228",
    "doubao-lite-4k-240328",
    "doubao-lite-4k-character-240515",
    "doubao-lite-4k-character-240828",
    "doubao-lite-4k-pretrain-character-240516",
    "doubao-pro-128k-240515",
    "doubao-pro-128k-240628",
    "doubao-pro-256k-241115",
    "doubao-pro-32k-240615",
    "doubao-pro-32k-240828",
    "doubao-pro-32k-241215",
    "doubao-pro-32k-browsing-240615",
    "doubao-pro-32k-browsing-240828",
    "doubao-pro-32k-browsing-241115",
    "doubao-pro-32k-character-240528",
    "doubao-pro-32k-character-240828",
    "doubao-pro-32k-character-241215",
    "doubao-pro-32k-functioncall-240515",
    "doubao-pro-32k-functioncall-240815",
    "doubao-pro-32k-functioncall-241028",
    "doubao-pro-32k-functioncall-preview",
    "doubao-pro-4k-240515",
    "doubao-pro-4k-browsing-240524",
    "doubao-pro-4k-character-240515",
    "doubao-pro-4k-character-240728",
    "doubao-pro-4k-functioncall-240515",
    "doubao-pro-4k-functioncall-240615",
    "doubao-seaweed-241128",
    "doubao-seed-1-6-250615",
    "doubao-seed-1-6-251015",
    "doubao-seed-1-6-flash-250615",
    "doubao-seed-1-6-flash-250715",
    "doubao-seed-1-6-flash-250828",
    "doubao-seed-1-6-lite-251015",
    "doubao-seed-1-6-thinking-250615",
    "doubao-seed-1-6-thinking-250715",
    "doubao-seed-1-6-vision-250815",
    "doubao-seed-1-8-251228",
    "doubao-seed-code-preview-251028",
    "doubao-seed-translation-250915",
    "doubao-seed3d-1-0-250928",
    "doubao-seedance-1-0-pro-250528",
    "doubao-seedance-1-0-pro-fast-251015",
    "doubao-seedance-1-5-pro-251215",
    "doubao-seededit-3-0-i2i-250628",
    "doubao-smart-router-250928",
    "doubao-vision-lite-32k-241015",
    "doubao-vision-pro-32k-241028",
    "glm-4-5-air-20250728",
    "kimi-k2-250711",
    "kimi-k2-250905",
    "kimi-k2-thinking-251104",
    "mistral-7b-instruct-v0.2",
    "qwen2-5-72b-20240919",
    "qwen3-0-6b-20250429",
    "qwen3-14b-20250429",
    "qwen3-32b-20250429",
    "qwen3-8b-20250429",
    "wan2-1-14b-flf2v-250417",
    "wan2-1-14b-i2v-250225",
    "wan2-1-14b-t2v-250225",
    "doubao-seedream-3-0-t2i-250415",
    "doubao-seedream-4-0-250828",
    "doubao-seedream-4-5-251128",
    "doubao-seedance-1-0-lite-i2v-250428",
    "doubao-seedance-1-0-lite-t2v-250428",
    "doubao-embedding-large-text-240915",
    "doubao-embedding-large-text-250515",
    "doubao-embedding-text-240515",
    "doubao-embedding-text-240715",
]



import asyncio
import unittest
from unittest.mock import patch

try:
    from mcp.server.fastmcp.exceptions import ToolError  # noqa: F401
except ModuleNotFoundError as e:
    raise unittest.SkipTest("missing dependency: mcp") from e


def _call_tool(server, name: str, args: dict) -> object:
    out = server.call_tool(name, args)
    if asyncio.iscoroutine(out):
        out = asyncio.run(out)
    if isinstance(out, tuple) and len(out) == 2 and isinstance(out[1], dict):
        return out[1]
    return out


class TestMcpModelKeywordFilter(unittest.TestCase):
    def test_list_available_models_filters_by_keyword(self) -> None:
        from gravtice.genai.mcp_server import build_server

        with patch(
            "gravtice.genai.mcp_server.Client.list_available_models",
            return_value=["gpt-4o-mini", "gpt-image-1"],
        ):
            server = build_server(host="127.0.0.1", port=7001, model_keywords=["image"])
            out = _call_tool(server, "list_available_models", {"provider": "openai"})

        self.assertIsInstance(out, dict)
        models = [m["model"] for m in out["models"]]
        self.assertEqual(models, ["openai:gpt-image-1"])

    def test_list_all_available_models_filters_by_keyword_or(self) -> None:
        from gravtice.genai.mcp_server import build_server

        all_models = [
            "openai:gpt-4o-mini",
            "openai:gpt-image-1",
            "tuzi-openai:gpt-4o-mini",
            "tuzi-openai:gpt-image-1",
        ]
        with patch(
            "gravtice.genai.mcp_server.Client.list_all_available_models",
            return_value=all_models,
        ):
            server = build_server(
                host="127.0.0.1", port=7001, model_keywords=["tuzi", "image"]
            )
            out = _call_tool(server, "list_all_available_models", {})

        self.assertIsInstance(out, dict)
        models = [m["model"] for m in out["models"]]
        self.assertEqual(
            models,
            [
                "openai:gpt-image-1",
                "tuzi-openai:gpt-4o-mini",
                "tuzi-openai:gpt-image-1",
            ],
        )

    def test_generate_rejects_disallowed_model(self) -> None:
        from gravtice.genai.mcp_server import build_server

        req_dict = {
            "model": "openai:gpt-4o-mini",
            "input": [{"role": "user", "content": [{"type": "text", "text": "x"}]}],
            "output": {"modalities": ["text"]},
        }
        with patch(
            "gravtice.genai.mcp_server.Client.generate",
            side_effect=AssertionError("should not call provider"),
        ):
            server = build_server(host="127.0.0.1", port=7001, model_keywords=["image"])
            with self.assertRaises(ToolError) as cm:
                _call_tool(server, "generate", {"request": req_dict})

        self.assertIn("model not allowed", str(cm.exception))



import os
import unittest
from unittest.mock import patch


class TestClientTimeout(unittest.TestCase):
    def test_generate_passes_request_timeout_ms(self) -> None:
        from gravtice.genai.client import Client
        from gravtice.genai.types import (
            Capability,
            GenerateParams,
            GenerateRequest,
            GenerateResponse,
            Message,
            OutputSpec,
            Part,
        )

        class DummyAdapter:
            def __init__(self) -> None:
                self.last_timeout_ms: int | None = None

            def capabilities(self, model_id: str) -> Capability:  # noqa: ARG002
                return Capability(
                    input_modalities={"text"},
                    output_modalities={"text"},
                    supports_stream=False,
                    supports_job=False,
                )

            def generate(self, request: GenerateRequest, *, stream: bool):  # noqa: ARG002
                self.last_timeout_ms = request.params.timeout_ms
                return GenerateResponse(
                    id="r1",
                    provider="dummy",
                    model=request.model,
                    status="completed",
                    output=[Message(role="assistant", content=[Part.from_text("ok")])],
                )

        client = Client()
        adapter = DummyAdapter()
        req = GenerateRequest(
            model="dummy:demo",
            input=[Message(role="user", content=[Part.from_text("hi")])],
            output=OutputSpec(modalities=["text"]),
            params=GenerateParams(timeout_ms=999),
        )

        with patch.object(client, "_adapter", return_value=adapter):
            client.generate(req)

        self.assertEqual(adapter.last_timeout_ms, 999)

    def test_generate_defaults_timeout_ms_from_env(self) -> None:
        from gravtice.genai.client import Client
        from gravtice.genai.types import (
            Capability,
            GenerateRequest,
            GenerateResponse,
            Message,
            OutputSpec,
            Part,
        )

        class DummyAdapter:
            def __init__(self) -> None:
                self.last_timeout_ms: int | None = None

            def capabilities(self, model_id: str) -> Capability:  # noqa: ARG002
                return Capability(
                    input_modalities={"text"},
                    output_modalities={"text"},
                    supports_stream=False,
                    supports_job=False,
                )

            def generate(self, request: GenerateRequest, *, stream: bool):  # noqa: ARG002
                self.last_timeout_ms = request.params.timeout_ms
                return GenerateResponse(
                    id="r1",
                    provider="dummy",
                    model=request.model,
                    status="completed",
                    output=[Message(role="assistant", content=[Part.from_text("ok")])],
                )

        with patch.dict(os.environ, {"GENAI_CALLING_TIMEOUT_MS": "12345"}, clear=False):
            client = Client()
            adapter = DummyAdapter()
            req = GenerateRequest(
                model="dummy:demo",
                input=[Message(role="user", content=[Part.from_text("hi")])],
                output=OutputSpec(modalities=["text"]),
            )
            with patch.object(client, "_adapter", return_value=adapter):
                client.generate(req)

        self.assertEqual(adapter.last_timeout_ms, 12345)

    def test_generate_stream_passes_request_timeout_ms(self) -> None:
        from gravtice.genai.client import Client
        from gravtice.genai.types import (
            Capability,
            GenerateEvent,
            GenerateParams,
            GenerateRequest,
            Message,
            OutputSpec,
            Part,
        )

        class DummyAdapter:
            def __init__(self) -> None:
                self.last_timeout_ms: int | None = None

            def capabilities(self, model_id: str) -> Capability:  # noqa: ARG002
                return Capability(
                    input_modalities={"text"},
                    output_modalities={"text"},
                    supports_stream=True,
                    supports_job=False,
                )

            def generate(self, request: GenerateRequest, *, stream: bool):
                if not stream:
                    raise AssertionError("expected stream=True")
                self.last_timeout_ms = request.params.timeout_ms
                return iter([GenerateEvent(type="done")])

        client = Client()
        adapter = DummyAdapter()
        req = GenerateRequest(
            model="dummy:demo",
            input=[Message(role="user", content=[Part.from_text("hi")])],
            output=OutputSpec(modalities=["text"]),
            params=GenerateParams(timeout_ms=456),
        )

        with patch.object(client, "_adapter", return_value=adapter):
            list(client.generate_stream(req))

        self.assertEqual(adapter.last_timeout_ms, 456)



from __future__ import annotations

import copy
from typing import Any, Literal

from ..types import Capability

from .model_catalog import MODEL_CATALOG
from .mode_overrides import CAPABILITY_OVERRIDES


def get_model_catalog() -> dict[str, list[str]]:
    return copy.deepcopy(MODEL_CATALOG)


def get_supported_providers() -> list[str]:
    return list(MODEL_CATALOG.keys())


def get_sdk_supported_models() -> list[dict[str, Any]]:
    """
    Return a JSON-friendly table of all curated models and their SDK-level capabilities.
    """
    out: list[dict[str, Any]] = []
    for provider, model_ids in MODEL_CATALOG.items():
        protocol = _default_protocol(provider)
        for model_id in model_ids:
            cap = _capability_for(
                provider=provider, protocol=protocol, model_id=model_id
            )
            cap = _apply_capability_overrides(
                provider=provider, model_id=model_id, cap=cap
            )
            category = _category_for_capability(cap)
            modes = _modes_for(cap)
            notes: list[str] = []
            if cap.supports_job:
                notes.append("可能返回 running(job)，需轮询/等待")
            out.append(
                {
                    "provider": provider,
                    "model_id": model_id,
                    "model": f"{provider}:{model_id}",
                    "category": category,
                    "protocol": protocol,
                    "modes": modes,
                    "input_modalities": sorted(cap.input_modalities),
                    "output_modalities": sorted(cap.output_modalities),
                    "supports_job": cap.supports_job,
                    "notes": notes,
                }
            )
    return out


def get_sdk_supported_models_for_provider(provider: str) -> list[dict[str, Any]]:
    """
    Return SDK-curated supported models for a single provider (same rows as `get_sdk_supported_models()`).

    Note: `provider` must match keys used in `MODEL_CATALOG` (e.g. "google", not "gemini").
    """
    p = provider.strip().lower()
    if not p:
        return []
    return [m for m in get_sdk_supported_models() if m.get("provider") == p]


def _default_protocol(provider: str) -> str:
    if provider in {"google", "tuzi-google"}:
        return "gemini"
    if provider in {"anthropic", "tuzi-anthropic"}:
        return "messages"
    if provider == "tuzi-web":
        return "multi"
    return "chat_completions"


def _category_for_capability(cap: Capability) -> str:
    out = set(cap.output_modalities)
    if out == {"embedding"}:
        return "embedding"
    if out == {"image"}:
        return "image"
    if out == {"video"}:
        return "video"
    if out == {"audio"}:
        return "audio"
    if (
        out == {"text"}
        and "audio" in set(cap.input_modalities)
        and not cap.supports_stream
    ):
        return "transcription"
    return "chat"


def _capability_for(*, provider: str, protocol: str, model_id: str) -> Capability:
    if provider in {"openai", "tuzi-openai"}:
        from ..providers import OpenAIAdapter

        return OpenAIAdapter(
            api_key="__demo__", provider_name=provider, chat_api=protocol
        ).capabilities(model_id)

    if provider in {"google", "tuzi-google"}:
        from ..providers import GeminiAdapter

        gemini_auth_mode: Literal["bearer", "query_key"] = (
            "bearer" if provider.startswith("tuzi-") else "query_key"
        )
        return GeminiAdapter(
            api_key="__demo__", provider_name=provider, auth_mode=gemini_auth_mode
        ).capabilities(model_id)

    if provider in {"anthropic", "tuzi-anthropic"}:
        from ..providers import AnthropicAdapter

        anthropic_auth_mode: Literal["bearer", "x-api-key"] = (
            "bearer" if provider.startswith("tuzi-") else "x-api-key"
        )
        return AnthropicAdapter(
            api_key="__demo__", provider_name=provider, auth_mode=anthropic_auth_mode
        ).capabilities(model_id)

    if provider == "volcengine":
        from ..providers import OpenAIAdapter, VolcengineAdapter

        volc = VolcengineAdapter(
            openai=OpenAIAdapter(
                api_key="__demo__",
                provider_name="volcengine",
                chat_api="chat_completions",
            )
        )
        return volc.capabilities(model_id)

    if provider == "aliyun":
        from ..providers import AliyunAdapter, OpenAIAdapter

        aliyun = AliyunAdapter(
            openai=OpenAIAdapter(
                api_key="__demo__", provider_name="aliyun", chat_api="chat_completions"
            )
        )
        return aliyun.capabilities(model_id)

    if provider == "tuzi-web":
        from ..providers import (
            AnthropicAdapter,
            GeminiAdapter,
            OpenAIAdapter,
            TuziAdapter,
        )

        return TuziAdapter(
            openai=OpenAIAdapter(
                api_key="__demo__", provider_name=provider, chat_api="chat_completions"
            ),
            gemini=GeminiAdapter(
                api_key="__demo__",
                base_url="https://api.tu-zi.com",
                provider_name=provider,
                auth_mode="bearer",
                supports_file_upload=False,
            ),
            anthropic=AnthropicAdapter(
                api_key="__demo__", provider_name=provider, auth_mode="bearer"
            ),
        ).capabilities(model_id)

    raise ValueError(f"unknown provider: {provider}")


def _apply_capability_overrides(
    *, provider: str, model_id: str, cap: Capability
) -> Capability:
    overrides = CAPABILITY_OVERRIDES.get(provider)
    if not overrides:
        return cap
    override = overrides.get(model_id)
    if not override:
        return cap
    supports_stream = cap.supports_stream
    if "supports_stream" in override:
        supports_stream = override["supports_stream"]
    supports_job = cap.supports_job
    if "supports_job" in override:
        supports_job = override["supports_job"]
    if supports_stream == cap.supports_stream and supports_job == cap.supports_job:
        return cap
    return Capability(
        input_modalities=set(cap.input_modalities),
        output_modalities=set(cap.output_modalities),
        supports_stream=supports_stream,
        supports_job=supports_job,
    )


def _modes_for(cap: Capability) -> list[str]:
    modes: list[str] = ["sync"]
    if cap.supports_stream:
        modes.append("stream")
    if cap.supports_job:
        modes.append("job")
    modes.append("async")
    return modes



import unittest


class TestCliProbeParsing(unittest.TestCase):
    def test_parse_probe_models_csv_and_dedup(self) -> None:
        from gravtice.genai.cli import _parse_probe_models

        out = _parse_probe_models("openai", "gpt-4o-mini, gpt-4o-mini , gpt-4")
        self.assertEqual(out, ["gpt-4o-mini", "gpt-4"])

    def test_parse_probe_models_accepts_provider_prefix(self) -> None:
        from gravtice.genai.cli import _parse_probe_models

        out = _parse_probe_models("openai", "openai:gpt-4o-mini,openai:gpt-4")
        self.assertEqual(out, ["gpt-4o-mini", "gpt-4"])

    def test_parse_probe_models_rejects_provider_mismatch(self) -> None:
        from gravtice.genai.cli import _parse_probe_models

        with self.assertRaises(SystemExit):
            _parse_probe_models("openai", "google:gemini-2.0-flash")



from __future__ import annotations

# Curated model catalog used by demos/CLI.
#
# Catalog only lists model ids per provider.
# Capabilities (and display categories) are inferred at runtime from Adapter.capabilities().

from .model_catalog_data.aliyun import MODELS as ALIYUN_MODELS
from .model_catalog_data.anthropic import MODELS as ANTHROPIC_MODELS
from .model_catalog_data.google import MODELS as GOOGLE_MODELS
from .model_catalog_data.openai import MODELS as OPENAI_MODELS
from .model_catalog_data.tuzi_anthropic import MODELS as TUZI_ANTHROPIC_MODELS
from .model_catalog_data.tuzi_google import MODELS as TUZI_GOOGLE_MODELS
from .model_catalog_data.tuzi_openai import MODELS as TUZI_OPENAI_MODELS
from .model_catalog_data.tuzi_web import MODELS as TUZI_WEB_MODELS
from .model_catalog_data.volcengine import MODELS as VOLCENGINE_MODELS


MODEL_CATALOG: dict[str, list[str]] = {
    "openai": OPENAI_MODELS,
    "anthropic": ANTHROPIC_MODELS,
    "google": GOOGLE_MODELS,
    "volcengine": VOLCENGINE_MODELS,
    "aliyun": ALIYUN_MODELS,
    "tuzi-web": TUZI_WEB_MODELS,
    "tuzi-openai": TUZI_OPENAI_MODELS,
    "tuzi-google": TUZI_GOOGLE_MODELS,
    "tuzi-anthropic": TUZI_ANTHROPIC_MODELS,
}


SUPPORTED_TESTS: dict[str, list[str]] = {
    "openai": [
        "chat",
        "tools",
        "transcription",
        "image",
        "video",
        "audio",
        "embedding",
    ],
    "anthropic": ["chat", "tools"],
    "google": [
        "chat",
        "tools",
        "transcription",
        "image",
        "video",
        "audio",
        "embedding",
    ],
    "volcengine": ["chat", "image", "video", "embedding"],
    "aliyun": ["chat", "transcription", "image", "video", "audio", "embedding"],
    "tuzi-web": [
        "chat",
        "tools",
        "transcription",
        "image",
        "video",
        "audio",
        "embedding",
    ],
    "tuzi-openai": [
        "chat",
        "tools",
        "transcription",
        "image",
        "video",
        "audio",
        "embedding",
    ],
    "tuzi-google": [
        "chat",
        "tools",
        "transcription",
        "image",
        "video",
        "audio",
        "embedding",
    ],
    "tuzi-anthropic": ["chat", "tools"],
}



from __future__ import annotations

from typing import Any

from .errors import invalid_request_error, not_supported_error

_GEMINI_RESPONSE_SCHEMA_TYPES = frozenset(
    {
        "TYPE_UNSPECIFIED",
        "STRING",
        "NUMBER",
        "INTEGER",
        "BOOLEAN",
        "ARRAY",
        "OBJECT",
    }
)


def reject_gemini_response_schema_dict(schema: dict[str, Any]) -> None:
    t = schema.get("type")
    if isinstance(t, str) and t in _GEMINI_RESPONSE_SCHEMA_TYPES:
        raise invalid_request_error(
            "output.text.json_schema must be JSON Schema (not Gemini responseSchema); "
            "pass a Python type/model or use provider_options.google.generationConfig.responseSchema"
        )


def python_type_to_json_schema(schema: Any) -> dict[str, Any]:
    try:
        from pydantic import TypeAdapter
    except ModuleNotFoundError as e:  # pragma: no cover
        raise not_supported_error(
            "pydantic is required for python-type output.text.json_schema"
        ) from e

    try:
        return TypeAdapter(schema).json_schema()
    except Exception:
        pass

    try:
        return TypeAdapter(type(schema)).json_schema()
    except Exception as e:
        raise invalid_request_error(
            "output.text.json_schema must be a JSON Schema object or a Python type supported by pydantic TypeAdapter"
        ) from e


def normalize_json_schema(schema: Any) -> dict[str, Any]:
    if isinstance(schema, dict):
        reject_gemini_response_schema_dict(schema)
        return schema
    return python_type_to_json_schema(schema)



from __future__ import annotations

import json
import os
from dataclasses import dataclass
from typing import Any, Iterator, Literal
from uuid import uuid4

from .._internal.capability_rules import claude_input_modalities
from .._internal.errors import (
    invalid_request_error,
    not_supported_error,
    provider_error,
)
from .._internal.http import download_to_tempfile, request_json, request_stream_json_sse
from ..types import (
    Capability,
    GenerateEvent,
    GenerateRequest,
    GenerateResponse,
    Message,
    Part,
    PartSourceBytes,
    PartSourcePath,
    PartSourceUrl,
    Usage,
    bytes_to_base64,
    detect_mime_type,
    file_to_bytes,
    normalize_reasoning_effort,
)


_ANTHROPIC_DEFAULT_BASE_URL = "https://api.anthropic.com"

_INLINE_BYTES_LIMIT = 20 * 1024 * 1024

_DEFAULT_VERSION = "2023-06-01"


_EFFORT_TO_THINKING_BUDGET_TOKENS: dict[str, int] = {
    "minimal": 1_024,
    "low": 1_024,
    "medium": 2_048,
    "high": 4_096,
    "xhigh": 8_192,
}


@dataclass(frozen=True, slots=True)
class AnthropicAdapter:
    api_key: str
    base_url: str = _ANTHROPIC_DEFAULT_BASE_URL
    provider_name: str = "anthropic"
    auth_mode: Literal["x-api-key", "bearer"] = "x-api-key"
    version: str = _DEFAULT_VERSION
    proxy_url: str | None = None

    def capabilities(self, model_id: str) -> Capability:
        mid = model_id.strip()
        if not mid:
            raise invalid_request_error("model_id must not be empty")
        return Capability(
            input_modalities=claude_input_modalities(mid),
            output_modalities={"text"},
            supports_stream=True,
            supports_job=False,
            supports_tools=True,
            supports_json_schema=False,
        )

    def list_models(self, *, timeout_ms: int | None = None) -> list[str]:
        """
        Fetch remote model ids via Anthropic GET /v1/models.
        """
        url = f"{self.base_url.rstrip('/')}/v1/models"
        obj = request_json(
            method="GET",
            url=url,
            headers=self._headers(),
            timeout_ms=timeout_ms,
            proxy_url=self.proxy_url,
        )
        data = obj.get("data")
        if not isinstance(data, list):
            return []
        out: list[str] = []
        for item in data:
            if not isinstance(item, dict):
                continue
            mid = item.get("id")
            if isinstance(mid, str) and mid:
                out.append(mid)
        return sorted(set(out))

    def generate(
        self, request: GenerateRequest, *, stream: bool
    ) -> GenerateResponse | Iterator[GenerateEvent]:
        if set(request.output.modalities) != {"text"}:
            raise not_supported_error("Anthropic only supports text output in this SDK")
        if request.output.text and (
            request.output.text.format != "text"
            or request.output.text.json_schema is not None
        ):
            raise not_supported_error(
                "Anthropic json output is not supported in this SDK"
            )
        if request.params.seed is not None:
            raise not_supported_error("Anthropic does not support seed in this SDK")
        if stream:
            return self._messages_stream(request)
        return self._messages(request)

    def _headers(self, request: GenerateRequest | None = None) -> dict[str, str]:
        headers: dict[str, str] = {
            "Content-Type": "application/json",
            "anthropic-version": self.version,
        }
        if self.auth_mode == "bearer":
            headers["Authorization"] = f"Bearer {self.api_key}"
        else:
            headers["x-api-key"] = self.api_key
        if request and request.params.idempotency_key:
            headers["Idempotency-Key"] = request.params.idempotency_key
        return headers

    def _messages(self, request: GenerateRequest) -> GenerateResponse:
        url = f"{self.base_url.rstrip('/')}/v1/messages"
        body = self._messages_body(request, stream=False)
        obj = request_json(
            method="POST",
            url=url,
            headers=self._headers(request),
            json_body=body,
            timeout_ms=request.params.timeout_ms,
            proxy_url=self.proxy_url,
        )
        return self._parse_message(obj, model_id=request.model_id())

    def _messages_stream(self, request: GenerateRequest) -> Iterator[GenerateEvent]:
        url = f"{self.base_url.rstrip('/')}/v1/messages"
        body = self._messages_body(request, stream=True)
        events = request_stream_json_sse(
            method="POST",
            url=url,
            headers=self._headers(request),
            json_body=body,
            timeout_ms=request.params.timeout_ms,
            proxy_url=self.proxy_url,
        )

        def _iter() -> Iterator[GenerateEvent]:
            for obj in events:
                if isinstance(obj.get("data"), dict):
                    obj = obj["data"]
                if not isinstance(obj, dict):
                    continue
                if obj.get("type") == "content_block_delta":
                    delta = obj.get("delta")
                    if isinstance(delta, dict) and delta.get("type") == "text_delta":
                        text = delta.get("text")
                        if isinstance(text, str) and text:
                            yield GenerateEvent(
                                type="output.text.delta", data={"delta": text}
                            )
            yield GenerateEvent(type="done", data={})

        return _iter()

    def _messages_body(
        self, request: GenerateRequest, *, stream: bool
    ) -> dict[str, Any]:
        model_id = request.model_id()
        max_tokens = _max_tokens(request)
        system = _extract_system_text(request)

        messages: list[dict[str, Any]] = []
        for m in request.input:
            if m.role == "system":
                continue
            if m.role not in {"user", "assistant", "tool"}:
                raise not_supported_error(f"Anthropic does not support role: {m.role}")
            if m.role == "user" and any(p.type == "tool_result" for p in m.content):
                raise invalid_request_error(
                    "tool_result parts must be sent as role='tool' for Anthropic"
                )
            if m.role == "user" and any(p.type == "tool_call" for p in m.content):
                raise invalid_request_error(
                    "tool_call parts are only allowed in assistant messages"
                )
            if m.role == "assistant" and any(
                p.type == "tool_result" for p in m.content
            ):
                raise invalid_request_error(
                    "tool_result parts must be sent as role='tool' for Anthropic"
                )
            if m.role == "tool" and any(p.type != "tool_result" for p in m.content):
                raise invalid_request_error(
                    "tool messages may only contain tool_result parts"
                )
            blocks = [
                _part_to_block(
                    p, timeout_ms=request.params.timeout_ms, proxy_url=self.proxy_url
                )
                for p in m.content
            ]
            role = "user" if m.role == "tool" else m.role
            messages.append({"role": role, "content": blocks})

        if not messages:
            raise invalid_request_error(
                "request.input must contain at least one non-system message"
            )

        body: dict[str, Any] = {
            "model": model_id,
            "max_tokens": max_tokens,
            "messages": messages,
            "stream": stream,
        }
        if system:
            body["system"] = system

        params = request.params
        if params.temperature is not None:
            body["temperature"] = params.temperature
        if params.top_p is not None:
            body["top_p"] = params.top_p
        if params.stop is not None:
            body["stop_sequences"] = params.stop
        thinking = _thinking_param(request, max_tokens=max_tokens)
        if thinking is not None:
            body["thinking"] = thinking

        if request.tools:
            tools: list[dict[str, Any]] = []
            for t in request.tools:
                name = t.name.strip()
                if not name:
                    raise invalid_request_error("tool.name must be non-empty")
                tool_obj: dict[str, Any] = {
                    "name": name,
                    "input_schema": t.parameters
                    if t.parameters is not None
                    else {"type": "object"},
                }
                if isinstance(t.description, str) and t.description.strip():
                    tool_obj["description"] = t.description.strip()
                tools.append(tool_obj)
            body["tools"] = tools

        if request.tool_choice is not None:
            choice = request.tool_choice.normalized()
            if choice.mode in {"required", "tool"} and not request.tools:
                raise invalid_request_error("tool_choice requires request.tools")
            if choice.mode == "required":
                body["tool_choice"] = {"type": "any"}
            elif choice.mode == "tool":
                body["tool_choice"] = {"type": "tool", "name": choice.name}
            else:
                body["tool_choice"] = {"type": choice.mode}

        opts = request.provider_options.get(self.provider_name)
        if isinstance(opts, dict):
            for k, v in opts.items():
                if k in body:
                    raise invalid_request_error(
                        f"provider_options cannot override body.{k}"
                    )
                body[k] = v
        return body

    def _parse_message(self, obj: dict[str, Any], *, model_id: str) -> GenerateResponse:
        if isinstance(obj.get("data"), dict):
            obj = obj["data"]
        content = obj.get("content")
        if not isinstance(content, list):
            raise provider_error("anthropic response missing content")
        parts: list[Part] = []
        for item in content:
            if not isinstance(item, dict):
                continue
            typ = item.get("type")
            if typ == "text":
                t = item.get("text")
                if isinstance(t, str):
                    parts.append(Part.from_text(t))
                continue
            if typ == "tool_use":
                tool_use_id = item.get("id")
                name = item.get("name")
                tool_input = item.get("input")
                if (
                    isinstance(tool_use_id, str)
                    and tool_use_id
                    and isinstance(name, str)
                    and name
                    and isinstance(tool_input, dict)
                ):
                    parts.append(
                        Part.tool_call(
                            tool_call_id=tool_use_id, name=name, arguments=tool_input
                        )
                    )

        usage_obj = obj.get("usage")
        usage = None
        if isinstance(usage_obj, dict):
            usage = Usage(
                input_tokens=usage_obj.get("input_tokens"),
                output_tokens=usage_obj.get("output_tokens"),
                total_tokens=usage_obj.get("input_tokens", 0)
                + usage_obj.get("output_tokens", 0)
                if isinstance(usage_obj.get("input_tokens"), int)
                and isinstance(usage_obj.get("output_tokens"), int)
                else None,
            )

        raw_id = obj.get("id")
        resp_id = raw_id if isinstance(raw_id, str) and raw_id else f"sdk_{uuid4().hex}"

        return GenerateResponse(
            id=resp_id,
            provider=self.provider_name,
            model=f"{self.provider_name}:{model_id}",
            status="completed",
            output=[
                Message(
                    role="assistant", content=parts if parts else [Part.from_text("")]
                )
            ],
            usage=usage,
        )


def _max_tokens(request: GenerateRequest) -> int:
    spec = request.output.text
    if spec and spec.max_output_tokens is not None:
        return max(1, int(spec.max_output_tokens))
    if request.params.max_output_tokens is not None:
        return max(1, int(request.params.max_output_tokens))
    return 1024


def _thinking_param(
    request: GenerateRequest, *, max_tokens: int
) -> dict[str, Any] | None:
    reasoning = request.params.reasoning
    if reasoning is None:
        return None
    if reasoning.effort is None:
        return None
    effort = normalize_reasoning_effort(reasoning.effort)
    if effort == "none":
        return None
    budget = _EFFORT_TO_THINKING_BUDGET_TOKENS[effort]
    if budget >= max_tokens:
        raise invalid_request_error(
            f"Anthropic thinking budget_tokens must be < max_tokens ({budget} >= {max_tokens}); "
            "increase output.text.max_output_tokens or params.max_output_tokens"
        )
    return {"type": "enabled", "budget_tokens": budget}


def _extract_system_text(request: GenerateRequest) -> str | None:
    chunks: list[str] = []
    for m in request.input:
        if m.role != "system":
            continue
        for p in m.content:
            if p.type != "text":
                raise not_supported_error("Anthropic system messages only support text")
            t = p.require_text().strip()
            if t:
                chunks.append(t)
    if not chunks:
        return None
    return "\n\n".join(chunks)


def _part_to_block(
    part: Part, *, timeout_ms: int | None, proxy_url: str | None
) -> dict[str, Any]:
    if part.type == "text":
        return {"type": "text", "text": part.require_text()}
    if part.type == "tool_call":
        tool_call_id = part.meta.get("tool_call_id")
        name = part.meta.get("name")
        arguments = part.meta.get("arguments")
        if not isinstance(tool_call_id, str) or not tool_call_id:
            raise invalid_request_error(
                "tool_call.meta.tool_call_id required for Anthropic tool_use"
            )
        if not isinstance(name, str) or not name.strip():
            raise invalid_request_error(
                "tool_call.meta.name must be a non-empty string"
            )
        if not isinstance(arguments, dict):
            raise invalid_request_error(
                "Anthropic tool_call.meta.arguments must be an object"
            )
        return {
            "type": "tool_use",
            "id": tool_call_id,
            "name": name.strip(),
            "input": arguments,
        }
    if part.type == "tool_result":
        tool_call_id = part.meta.get("tool_call_id")
        name = part.meta.get("name")
        result = part.meta.get("result")
        is_error = part.meta.get("is_error")
        if not isinstance(tool_call_id, str) or not tool_call_id:
            raise invalid_request_error(
                "tool_result.meta.tool_call_id required for Anthropic tool_result"
            )
        if not isinstance(name, str) or not name.strip():
            raise invalid_request_error(
                "tool_result.meta.name must be a non-empty string"
            )
        if is_error is not None and not isinstance(is_error, bool):
            raise invalid_request_error("tool_result.meta.is_error must be a bool")
        out = (
            result
            if isinstance(result, str)
            else json.dumps(result, ensure_ascii=False, separators=(",", ":"))
        )
        block: dict[str, Any] = {
            "type": "tool_result",
            "tool_use_id": tool_call_id,
            "content": out,
        }
        if is_error is not None:
            block["is_error"] = is_error
        return block
    if part.type == "image":
        source = part.require_source()
        mime_type = part.mime_type
        if mime_type is None and isinstance(source, PartSourcePath):
            mime_type = detect_mime_type(source.path)
        if not mime_type or not mime_type.startswith("image/"):
            raise invalid_request_error("anthropic image requires image/* mime_type")

        if isinstance(source, PartSourceUrl):
            tmp = download_to_tempfile(
                url=source.url,
                timeout_ms=timeout_ms,
                max_bytes=_INLINE_BYTES_LIMIT,
                proxy_url=proxy_url,
            )
            try:
                data = file_to_bytes(tmp, _INLINE_BYTES_LIMIT)
            finally:
                try:
                    os.unlink(tmp)
                except OSError:
                    pass
            data_b64 = bytes_to_base64(data)
        elif isinstance(source, PartSourcePath):
            data = file_to_bytes(source.path, _INLINE_BYTES_LIMIT)
            data_b64 = bytes_to_base64(data)
        elif isinstance(source, PartSourceBytes) and source.encoding == "base64":
            raw_b64 = source.data
            if not isinstance(raw_b64, str) or not raw_b64:
                raise invalid_request_error("image base64 data must be non-empty")
            data_b64 = raw_b64
        else:
            assert isinstance(source, PartSourceBytes)
            raw = source.data
            if not isinstance(raw, bytes):
                raise invalid_request_error("image bytes data must be bytes")
            data = raw
            if len(data) > _INLINE_BYTES_LIMIT:
                raise not_supported_error(
                    f"inline bytes too large ({len(data)} > {_INLINE_BYTES_LIMIT})"
                )
            data_b64 = bytes_to_base64(data)

        return {
            "type": "image",
            "source": {"type": "base64", "media_type": mime_type, "data": data_b64},
        }
    raise not_supported_error(f"Anthropic does not support part type: {part.type}")



import tempfile
import unittest
from unittest.mock import patch


class TestClientProtectedUrlArtifacts(unittest.TestCase):
    def test_externalize_protected_url_to_artifact(self) -> None:
        from gravtice.genai.client import Client
        from gravtice.genai.types import GenerateResponse, Message, Part, PartSourceUrl

        class DummyStore:
            def __init__(self) -> None:
                self.items: list[tuple[bytes, str | None]] = []

            def put(self, data: bytes, mime_type: str | None) -> str | None:
                self.items.append((data, mime_type))
                return "a1"

            def url(self, artifact_id: str) -> str:
                return f"http://artifact/{artifact_id}"

        class DummyAdapter:
            base_url = "https://generativelanguage.googleapis.com"

            def _download_headers(self) -> dict[str, str]:
                return {"x-goog-api-key": "k"}

        def fake_download_to_tempfile(**_kwargs) -> str:
            with tempfile.NamedTemporaryFile(
                prefix="genaisdk-test-", delete=False
            ) as f:
                f.write(b"video-bytes")
                return f.name

        client = Client(artifact_store=DummyStore())
        resp = GenerateResponse(
            id="r1",
            provider="google",
            model="google:models/veo-2.0-generate-001",
            status="completed",
            output=[
                Message(
                    role="assistant",
                    content=[
                        Part(
                            type="video",
                            mime_type="video/mp4",
                            source=PartSourceUrl(
                                url="https://generativelanguage.googleapis.com/v1beta/files/f1:download?alt=media"
                            ),
                        )
                    ],
                )
            ],
        )

        with patch(
            "gravtice.genai.client.download_to_tempfile",
            side_effect=fake_download_to_tempfile,
        ):
            out = client._externalize_protected_url_parts(
                resp, adapter=DummyAdapter(), timeout_ms=1
            )

        part = out.output[0].content[0]
        assert part.source is not None
        self.assertEqual(part.source.kind, "url")
        self.assertEqual(part.source.url, "http://artifact/a1")

    def test_download_to_file_adds_headers_only_for_same_host(self) -> None:
        from gravtice.genai.client import Client

        class DummyAdapter:
            base_url = "https://generativelanguage.googleapis.com"

            def _download_headers(self) -> dict[str, str]:
                return {"x-goog-api-key": "k"}

        client = Client()
        captured: dict[str, object] = {}

        def fake_download_to_file(**kwargs) -> None:
            captured.update(kwargs)

        with (
            patch.object(client, "_adapter", return_value=DummyAdapter()),
            patch(
                "gravtice.genai.client._download_to_file",
                side_effect=fake_download_to_file,
            ),
        ):
            client.download_to_file(
                provider="google",
                url="https://generativelanguage.googleapis.com/v1beta/files/f1:download?alt=media",
                output_path="out.mp4",
            )
        self.assertEqual(captured.get("headers"), {"x-goog-api-key": "k"})

        captured.clear()
        with (
            patch.object(client, "_adapter", return_value=DummyAdapter()),
            patch(
                "gravtice.genai.client._download_to_file",
                side_effect=fake_download_to_file,
            ),
        ):
            client.download_to_file(
                provider="google",
                url="https://example.com/video.mp4",
                output_path="out.mp4",
            )
        self.assertIsNone(captured.get("headers"))



import unittest


class TestMcpServerTransports(unittest.TestCase):
    def test_build_http_app_exposes_streamable_and_sse(self) -> None:
        try:
            from mcp.server.fastmcp import FastMCP  # noqa: F401
        except ModuleNotFoundError:
            self.skipTest("missing dependency: mcp")

        try:
            from starlette.routing import Mount, Route  # noqa: F401
        except ModuleNotFoundError:
            self.skipTest("missing dependency: starlette")

        from gravtice.genai.mcp_server import build_http_app, build_server

        server = build_server(host="127.0.0.1", port=7001)
        app = build_http_app(server)

        paths = {getattr(route, "path", None) for route in app.router.routes}
        self.assertIn(server.settings.streamable_http_path, paths)
        self.assertIn(server.settings.sse_path, paths)
        self.assertIn(server.settings.message_path.rstrip("/"), paths)



from __future__ import annotations

MODELS: list[str] = [
    "gemini-2.5-flash",
    "gemini-2.5-flash-lite",
    "gemini-2.5-flash-nothink",
    "gemini-2.5-flash-search",
    "gemini-2.5-pro",
    "gemini-2.5-pro-search",
    "gemini-3-flash-preview",
    "gemini-3-pro-preview",
    "gemini-3-pro-preview-search",
    "gemini-2.0-flash-exp-image-generation",
    "gemini-2.0-flash-preview-image-generation",
    "gemini-2.5-flash-image",
    "gemini-3-pro-image-preview",
    "gemini-embedding-001",
    "gemini-embedding-exp-03-07",
]



import unittest
from unittest.mock import patch


class TestOpenAIResponsesStreaming(unittest.TestCase):
    def _request(self):
        from gravtice.genai.types import GenerateRequest, Message, OutputSpec, Part

        return GenerateRequest(
            model="openai:gpt-5.2",
            input=[Message(role="user", content=[Part.from_text("hi")])],
            output=OutputSpec(modalities=["text"]),
        )

    def test_error_event_raises_provider_error(self) -> None:
        from gravtice.genai import GenAIError
        from gravtice.genai.providers.openai import OpenAIAdapter

        req = self._request()
        adapter = OpenAIAdapter(
            api_key="__demo__",
            base_url="https://example.invalid/v1",
            provider_name="openai",
            chat_api="responses",
        )

        with patch(
            "gravtice.genai.providers.openai.request_stream_json_sse"
        ) as request_stream_json_sse:
            request_stream_json_sse.return_value = iter(
                [
                    {"type": "response.output_text.delta", "delta": "hi"},
                    {
                        "type": "error",
                        "code": "ERR_SOMETHING",
                        "message": "Something went wrong",
                        "param": None,
                        "sequence_number": 1,
                    },
                ]
            )

            it = adapter.generate(req, stream=True)
            ev = next(it)
            self.assertEqual(ev.type, "output.text.delta")
            self.assertEqual(ev.data.get("delta"), "hi")

            with self.assertRaises(GenAIError) as cm:
                next(it)
            self.assertEqual(cm.exception.info.type, "ProviderError")
            self.assertEqual(cm.exception.info.provider_code, "ERR_SOMETHING")
            self.assertIn("Something went wrong", str(cm.exception))

    def test_failed_event_raises_provider_error(self) -> None:
        from gravtice.genai import GenAIError
        from gravtice.genai.providers.openai import OpenAIAdapter

        req = self._request()
        adapter = OpenAIAdapter(
            api_key="__demo__",
            base_url="https://example.invalid/v1",
            provider_name="openai",
            chat_api="responses",
        )

        with patch(
            "gravtice.genai.providers.openai.request_stream_json_sse"
        ) as request_stream_json_sse:
            request_stream_json_sse.return_value = iter(
                [
                    {
                        "type": "response.failed",
                        "sequence_number": 1,
                        "response": {
                            "error": {
                                "code": "server_error",
                                "message": "The model failed to generate a response.",
                            }
                        },
                    }
                ]
            )

            it = adapter.generate(req, stream=True)
            with self.assertRaises(GenAIError) as cm:
                next(it)
            self.assertEqual(cm.exception.info.type, "ProviderError")
            self.assertEqual(cm.exception.info.provider_code, "server_error")
            self.assertIn("failed to generate", str(cm.exception))

    def test_incomplete_event_raises_provider_error(self) -> None:
        from gravtice.genai import GenAIError
        from gravtice.genai.providers.openai import OpenAIAdapter

        req = self._request()
        adapter = OpenAIAdapter(
            api_key="__demo__",
            base_url="https://example.invalid/v1",
            provider_name="openai",
            chat_api="responses",
        )

        with patch(
            "gravtice.genai.providers.openai.request_stream_json_sse"
        ) as request_stream_json_sse:
            request_stream_json_sse.return_value = iter(
                [
                    {
                        "type": "response.incomplete",
                        "sequence_number": 1,
                        "response": {
                            "incomplete_details": {"reason": "max_output_tokens"}
                        },
                    }
                ]
            )

            it = adapter.generate(req, stream=True)
            with self.assertRaises(GenAIError) as cm:
                next(it)
            self.assertEqual(cm.exception.info.type, "ProviderError")
            self.assertEqual(cm.exception.info.provider_code, None)
            self.assertIn("incomplete", str(cm.exception))
            self.assertIn("max_output_tokens", str(cm.exception))



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class ErrorInfo:
    type: str
    message: str
    provider_code: str | None = None
    retryable: bool = False


class GenAIError(RuntimeError):
    def __init__(self, info: ErrorInfo):
        super().__init__(info.message)
        self.info = info


def auth_error(message: str, provider_code: str | None = None) -> GenAIError:
    return GenAIError(
        ErrorInfo(type="AuthError", message=message, provider_code=provider_code)
    )


def rate_limit_error(message: str, provider_code: str | None = None) -> GenAIError:
    return GenAIError(
        ErrorInfo(
            type="RateLimitError",
            message=message,
            provider_code=provider_code,
            retryable=True,
        )
    )


def invalid_request_error(message: str, provider_code: str | None = None) -> GenAIError:
    return GenAIError(
        ErrorInfo(
            type="InvalidRequestError", message=message, provider_code=provider_code
        )
    )


def not_supported_error(message: str) -> GenAIError:
    return GenAIError(ErrorInfo(type="NotSupportedError", message=message))


def timeout_error(message: str) -> GenAIError:
    return GenAIError(ErrorInfo(type="TimeoutError", message=message, retryable=True))


def provider_error(
    message: str, provider_code: str | None = None, retryable: bool = False
) -> GenAIError:
    return GenAIError(
        ErrorInfo(
            type="ProviderError",
            message=message,
            provider_code=provider_code,
            retryable=retryable,
        )
    )
